# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

"""
This module provides functions for creating CommandLine objects from
dictionaries and vice versa.
"""

from collections import OrderedDict

from .errors import CrazyError, CrazyTypeError
from .cli import CommandLine, ExtendedBool
from .str_utils import validate_prog
from . import compat


_INHERIT = ExtendedBool.INHERIT


def _validate_keys(dictionary, allowed_keys):
    for key in dictionary.keys():
        if key not in allowed_keys:
            raise CrazyError(f'Unknown key: {key}')


def dictionary_to_commandline(dictionary, prog=None):
    '''Convert a single dictionary to a cli.CommandLine object.'''

    _validate_keys(dictionary,
        ['prog', 'help', 'aliases', 'wraps',
         'abbreviate_commands', 'abbreviate_options',
         'inherit_options', 'options', 'positionals'])

    options = dictionary.get('options', [])
    if not isinstance(options, list):
        raise CrazyTypeError('options', 'list', options)

    positionals = dictionary.get('positionals', [])
    if not isinstance(positionals, list):
        raise CrazyTypeError('positionals', 'list', positionals)

    commandline = CommandLine(
        prog or dictionary['prog'],
        parent              = None,
        help                = dictionary.get('help', None),
        aliases             = dictionary.get('aliases', []),
        wraps               = dictionary.get('wraps', None),
        abbreviate_commands = dictionary.get('abbreviate_commands', _INHERIT),
        abbreviate_options  = dictionary.get('abbreviate_options', _INHERIT),
        inherit_options     = dictionary.get('inherit_options', _INHERIT))

    for option in options:
        _validate_keys(option,
            ['option_strings', 'metavar', 'help', 'optional_arg', 'groups',
             'repeatable', 'final', 'hidden', 'complete', 'nosort',
             'when', 'capture', 'long_opt_arg_sep'])

        commandline.add_option(
            option.get('option_strings',  None),
            metavar       = option.get('metavar',      None),
            help          = option.get('help',         None),
            optional_arg  = option.get('optional_arg', False),
            groups        = option.get('groups',       None),
            repeatable    = option.get('repeatable',   _INHERIT),
            final         = option.get('final',        False),
            hidden        = option.get('hidden',       False),
            complete      = option.get('complete',     None),
            nosort        = option.get('nosort',       False),
            when          = option.get('when',         None),
            capture       = option.get('capture',      None),
            long_opt_arg_sep = option.get('long_opt_arg_sep', _INHERIT)
        )

    for positional in positionals:
        _validate_keys(positional,
            ['number', 'metavar', 'help', 'repeatable', 'complete', 'nosort',
             'when', 'capture'])

        commandline.add_positional(
            positional.get('number', None),
            metavar    = positional.get('metavar',    None),
            help       = positional.get('help',       None),
            repeatable = positional.get('repeatable', False),
            complete   = positional.get('complete',   None),
            nosort     = positional.get('nosort',     False),
            when       = positional.get('when',       None),
            capture    = positional.get('capture',    None))

    return commandline


def _check_prog_in_dictionary(dictionary):
    if 'prog' not in dictionary:
        raise CrazyError('Missing `prog` field')

    if not isinstance(dictionary['prog'], str):
        raise CrazyTypeError('prog', 'str', dictionary["prog"])

    try:
        validate_prog(dictionary['prog'])
    except CrazyError as e:
        raise CrazyError(f'prog: {e}') from e


def _get_commandline_by_path(root, path):
    current = root

    for i, name in enumerate(path):
        path_str = ' '.join(path[0:i + 1])

        if not current.get_subcommands():
            raise CrazyError(f"Missing definition of program `{path_str}`")

        current = current.get_subcommands().get_subcommand_by_name(name)
        if not current:
            raise CrazyError(f"Missing definition of program `{path_str}`")

    return current


def replace_defines_in_documents(documents):
    '''Replaces defines in a list of documents.

    Exmaple:
        prog: '%defines%'
        my_completer: ['choices', ['foo', 'bar']]
        ---
        prog: 'example'
        options:
          - option_strings: ['-o']
            complete: 'my_completer'
    '''

    defines = None

    for document in documents:
        if not isinstance(document, dict):
            continue

        if 'prog' in document and document['prog'] == '%defines%':
            defines = document
            break

    if defines is None:
        return documents

    documents.remove(defines)
    defines.pop('prog')

    def replace_defines(obj):
        if isinstance(obj, str):
            return defines.get(obj, obj)
        if isinstance(obj, list):
            return [replace_defines(sub) for sub in obj]
        if isinstance(obj, dict):
            return {key: replace_defines(val) for key, val in obj.items()}
        return obj

    new = []
    for document in documents:
        new.append(replace_defines(document))
    return new


def dictionaries_to_commandline(dictionaries):
    '''Convert a list of dictionaries to a cli.CommandLine object.'''

    dictionaries = replace_defines_in_documents(dictionaries)
    compat.fix_commandline_dictionaries(dictionaries)

    root = CommandLine('root')

    for dictionary in dictionaries:
        _check_prog_in_dictionary(dictionary)

        path = dictionary['prog'].split()
        progname = path.pop()

        cmdline = _get_commandline_by_path(root, path)

        subcommands = cmdline.get_subcommands()
        if not subcommands:
            subcommands = cmdline.add_subcommands()

        new_cmdline = dictionary_to_commandline(dictionary, progname)

        try:
            subcommands.add_commandline_object(new_cmdline)
        except CrazyError as e:
            raise CrazyError(f"Multiple definition of program `{progname}`") from e

    if not root.get_subcommands():
        raise CrazyError("No programs defined")

    if len(root.get_subcommands().subcommands) > 1:
        progs = [c.prog for c in root.get_subcommands().subcommands]
        raise CrazyError(f"Too many main programs defined: {progs}")

    cmdline = root.get_subcommands().subcommands[0]
    cmdline.parent = None
    return cmdline


def option_to_dictionary(self):
    '''Convert a cli.Option object to a dictionary.'''

    r = OrderedDict()

    r['option_strings'] = self.option_strings

    if self.metavar is not None:
        r['metavar'] = self.metavar

    if self.help is not None:
        r['help'] = self.help

    if self.optional_arg is not False:
        r['optional_arg'] = self.optional_arg

    if self.groups is not None:
        r['groups'] = self.groups

    if self.repeatable != ExtendedBool.INHERIT:
        r['repeatable'] = self.repeatable

    if self.final is True:
        r['final'] = self.final

    if self.hidden is True:
        r['hidden'] = self.hidden

    if self.complete:
        r['complete'] = self.complete

    if self.nosort is True:
        r['nosort'] = self.nosort

    if self.when is not None:
        r['when'] = self.when

    if self.capture is not None:
        r['capture'] = self.capture

    if self.long_opt_arg_sep != ExtendedBool.INHERIT:
        r['long_opt_arg_sep'] = self.long_opt_arg_sep

    return r


def positional_to_dictionary(self):
    '''Convert a cli.Positional object to a dictionary.'''

    r = OrderedDict()

    r['number'] = self.number

    if self.metavar is not None:
        r['metavar'] = self.metavar

    if self.help is not None:
        r['help'] = self.help

    if self.repeatable is not False:
        r['repeatable'] = self.repeatable

    if self.complete:
        r['complete'] = self.complete

    if self.nosort is True:
        r['nosort'] = self.nosort

    if self.when is not None:
        r['when'] = self.when

    if self.capture is not None:
        r['capture'] = self.capture

    return r


def commandline_to_dictionary(commandline):
    '''Convert a cli.CommandLine object to a dictionary.'''

    r = OrderedDict()

    r['prog'] = commandline.get_command_path()

    if commandline.aliases:
        r['aliases'] = commandline.aliases

    if commandline.help is not None:
        r['help'] = commandline.help

    if commandline.wraps is not None:
        r['wraps'] = commandline.wraps

    if commandline.abbreviate_commands != ExtendedBool.INHERIT:
        r['abbreviate_commands'] = commandline.abbreviate_commands

    if commandline.abbreviate_options != ExtendedBool.INHERIT:
        r['abbreviate_options'] = commandline.abbreviate_options

    if commandline.inherit_options != ExtendedBool.INHERIT:
        r['inherit_options'] = commandline.inherit_options

    if commandline.options:
        r['options'] = options = []
        for option in commandline.options:
            options.append(option_to_dictionary(option))

    if commandline.positionals:
        r['positionals'] = positionals = []
        for positional in commandline.positionals:
            positionals.append(positional_to_dictionary(positional))

    return r


def commandline_to_dictionaries(commandline):
    '''Convert a cli.CommandLine object to dictionaries.'''

    dictionaries = []

    commandline.visit_commandlines(
        lambda c: dictionaries.append(commandline_to_dictionary(c)))

    return dictionaries



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Functions that are used in the generation process.'''

from .errors import CrazyError
from . import completion_validator
from . import cli
from . import config as _config
from . import when


class GenerationContext:
    '''Holds global configuration and helpers used during code generation.'''

    # pylint: disable=too-few-public-methods

    def __init__(self, config, helpers):
        self.config = config
        self.helpers = helpers

    def get_option_context(self, commandline, option):
        '''Return OptionGenerationContext based on current context.'''

        return OptionGenerationContext(
            self.config,
            self.helpers,
            commandline,
            option
        )


class OptionGenerationContext(GenerationContext):
    '''Extends GenerationContext; holds commandline and option objects.'''

    # pylint: disable=too-few-public-methods

    def __init__(self, config, helpers, commandline, option):
        super().__init__(config, helpers)
        self.commandline = commandline
        self.option = option


def _apply_config(commandline, config):
    '''Applies configuration settings to a command line object.

    If a setting in the CommandLine or Option object is set to
    ExtendedBool.INHERIT, it will be overridden by the corresponding setting
    from the config object.

    Args:
        commandline (CommandLine):
            The command line object to apply the configuration to.

        config (Config):
            The configuration object containing the settings to apply.

    Returns:
        None
    '''
    assert isinstance(commandline, cli.CommandLine)
    assert isinstance(config, _config.Config)

    if commandline.abbreviate_commands == cli.ExtendedBool.INHERIT:
        commandline.abbreviate_commands = config.abbreviate_commands

    if commandline.abbreviate_options == cli.ExtendedBool.INHERIT:
        commandline.abbreviate_options = config.abbreviate_options

    if commandline.inherit_options == cli.ExtendedBool.INHERIT:
        commandline.inherit_options = config.inherit_options

    for option in commandline.options:
        if option.repeatable == cli.ExtendedBool.INHERIT:
            option.repeatable = config.repeatable_options

        if option.long_opt_arg_sep == cli.ExtendedBool.INHERIT:
            option.long_opt_arg_sep = config.long_opt_arg_sep

        if config.disabled_hidden:
            option.hidden = False

        if config.disabled_final:
            option.final = False

        if config.disabled_groups:
            option.groups = []

        if config.disabled_repeatable:
            option.repeatable = True

        if config.disabled_when:
            option.when = None

    for positional in commandline.positionals:
        if config.disabled_when:
            positional.when = None


def _add_parsed_when(commandline):
    for option in commandline.options:
        if option.when:
            try:
                option.when_parsed = when.parse_when(option.when)
            except CrazyError as e:
                raise CrazyError('%s: %s: when: %s: %s' % (
                    commandline.get_command_path(),
                    option.get_option_strings_key('|'),
                    option.when,
                    e)) from e
        else:
            option.when_parsed = None

    for positional in commandline.positionals:
        if positional.when:
            try:
                positional.when_parsed = when.parse_when(positional.when)
            except CrazyError as e:
                raise CrazyError('%s: %s: %s: when: %s: %s' % (
                    commandline.get_command_path(),
                    positional.number,
                    positional.metavar,
                    positional.when,
                    e)) from e
        else:
            positional.when_parsed = None


def enhance_commandline(commandline, config):
    '''Enhance commandline.

    - Make a copy of commandline
    - Apply configuration to it
    - Add `when_parsed` attribute
    '''

    commandline = commandline.copy()
    commandline.visit_commandlines(lambda c: _apply_config(c, config))
    commandline.visit_commandlines(_add_parsed_when)
    completion_validator.validate_commandlines(commandline)
    return commandline


def visit_commandlines(completion_class, ctxt, commandline):
    '''Visit commandlines with a completer class.'''

    result = []

    def _call_generator(commandline):
        result.append(completion_class(ctxt, commandline))

    commandline.visit_commandlines(_call_generator)

    return result



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains code for completing arguments in Fish.'''

from . import shell
from .pattern import bash_glob_to_regex
from .type_utils import is_dict_type
from .str_utils import indent, join_with_wrap
from .utils import get_query_option_strings, get_defined_option_types


CHOICES_INLINE_THRESHOLD = 80
# The `choices` command can in Fish be expressed inline in a complete command
# like this:
#   complete -c program -a 'foo bar baz'
# or:
#   complete -c program -a 'foo\t"Foo value" bar\t"Bar value" baz\t"Baz value"'
#
# This variable defines how big this string can get before a function
# is used instead.


class FishCompletionBase:
    '''Base class for Fish completions.'''

    def __init__(self, ctxt):
        self.ctxt = ctxt

    def get_args(self):
        '''Return a list of arguments to be appended to the `complete`
        command in Fish.

        The returned arguments should be in raw form, without any escaping.
        Escaping will be handled at a later stage.
        '''
        raise NotImplementedError

    def get_code(self):
        '''Return the code that can be used for completing an argument.'''
        raise NotImplementedError

    def get_function(self):
        '''Return a function that runs the code.'''
        func = self.ctxt.helpers.add_dynamic_func(self.ctxt, self.get_code())
        return func


class FishCompleteNone(FishCompletionBase):
    '''Class for completing an argument without a completer.'''

    def get_args(self):
        return ['-f']

    def get_code(self):
        return ''

    def get_function(self):
        return 'true'


class FishCompletionCommand(FishCompletionBase):
    '''Class for executing a command.'''

    def __init__(self, ctxt, args):
        super().__init__(ctxt)
        self.args = args

    def get_code(self):
        return shell.join_quoted(self.args)

    def get_args(self):
        command = shell.join_quoted(self.args)
        return ['-f', '-a', '(%s)' % command]

    def get_function(self):
        if len(self.args) == 1:
            return self.args[0]

        func = self.ctxt.helpers.add_dynamic_func(self.ctxt, self.get_code())
        return func


class FishCompletionRawCommand(FishCompletionBase):
    '''Class for executing a command (without any escaping)'''

    def __init__(self, ctxt, command):
        super().__init__(ctxt)
        self.command = command

    def get_code(self):
        return self.command

    def get_args(self):
        return ['-f', '-a', '(%s)' % self.command]


class FishCompleteChoices(FishCompletionBase):
    '''Class for completing choices.'''

    def __init__(self, ctxt, choices):
        super().__init__(ctxt)
        self.choices = choices

    def _get_inline_for_list(self):
        return ' '.join(shell.quote(str(c)) for c in self.choices)

    def _get_inline_for_dict(self):
        def str0(s):
            return str(s) if s is not None else ''

        stringified = {str(item): str0(desc) for item, desc in self.choices.items()}
        q = shell.quote
        r = ['%s\\t%s' % (q(item), q(desc)) for item, desc in stringified.items()]
        return ' '.join(r)

    def _get_code_for_list(self):
        code = "printf '%s\\n' \\\n"
        quoted = [shell.quote(str(item)) for item in self.choices]
        line_length = self.ctxt.config.line_length - 2
        code += indent(join_with_wrap(' ', ' \\\n', line_length, quoted), 2)
        return code.rstrip(' \\\n')

    def _get_code_for_dict(self):
        code = "printf '%s\\t%s\\n' \\\n"
        for item, desc in self.choices.items():
            if desc is None:
                desc = ''
            code += '  %s %s \\\n' % (shell.quote(str(item)), shell.quote(str(desc)))
        return code.rstrip(' \\\n')

    def get_args(self):
        if is_dict_type(self.choices):
            arg = self._get_inline_for_dict()
        else:
            arg = self._get_inline_for_list()

        if len(arg) <= CHOICES_INLINE_THRESHOLD:
            return ['-f', '-a', arg]

        func = self.get_function()
        return ['-f', '-a', '(%s)' % func]

    def get_code(self):
        if is_dict_type(self.choices):
            return self._get_code_for_dict()
        return self._get_code_for_list()


def _get_extension_regex(extensions, fuzzy):
    patterns = []

    for extension in extensions:
        pattern = ''
        for c in extension:
            if c.isalpha():
                pattern += '[%s%s]' % (c.lower(), c.upper())
            elif c in ('.', '+'):
                pattern += f'\\{c}'
            else:
                pattern += c

        if fuzzy:
            pattern += '.*'

        patterns.append(pattern)

    return '|'.join(f'(.*\\.{pattern})' for pattern in patterns)


class FishCompleteFile(FishCompletionBase):
    '''Class for completing files.'''

    def __init__(self, ctxt, opts):
        super().__init__(ctxt)

        fuzzy = False
        directory = None
        extensions = None
        ignore_globs = None
        self.args = []

        if opts:
            fuzzy = opts.get('fuzzy', False)
            directory = opts.get('directory', None)
            extensions = opts.get('extensions', None)
            ignore_globs = opts.get('ignore_globs', None)

        if directory:
            self.args.extend(['-C', directory])

        if extensions:
            ctxt.helpers.use_function('filedir', 'regex')
            self.args.extend(['-r', _get_extension_regex(extensions, fuzzy)])

        if ignore_globs:
            ctxt.helpers.use_function('filedir', 'regex_ignore')
            patterns = map(bash_glob_to_regex, ignore_globs)
            patterns = [f'({p})' for p in patterns]
            self.args.extend(['-i', '|'.join(patterns)])

    def get_args(self):
        if len(self.args) == 0:
            return ['-F']

        func = self.ctxt.helpers.use_function('filedir')
        return FishCompletionCommand(self.ctxt, [func] + self.args).get_args()

    def get_code(self):
        func = self.ctxt.helpers.use_function('filedir')
        return FishCompletionCommand(self.ctxt, [func] + self.args).get_code()

    def get_function(self):
        func = self.ctxt.helpers.use_function('filedir')
        return FishCompletionCommand(self.ctxt, [func] + self.args).get_function()


class FishCompleteDirectory(FishCompletionCommand):
    '''Class for completing directories.'''

    def __init__(self, ctxt, trace, opts):
        directory = None if opts is None else opts.get('directory', None)

        # __fish_complete_directories does not respect __fish_stripprefix
        # which is used inside list/key_value_list/prefix
        use_filedir = (
            'list' in trace or
            'key_value_list' in trace or
            'prefix' in trace
        )

        if directory is not None:
            func = ctxt.helpers.use_function('filedir')
            super().__init__(ctxt, [func, '-D', '-C', directory])
        elif use_filedir:
            func = ctxt.helpers.use_function('filedir')
            super().__init__(ctxt, [func, '-D'])
        else:
            super().__init__(ctxt, ['__fish_complete_directories'])


class FishCompleteValueList(FishCompletionCommand):
    '''Class for completing a list of values.'''

    def __init__(self, ctxt, opts):
        separator = opts.get('separator', ',')
        duplicates = opts.get('duplicates', False)
        values = opts['values']

        if is_dict_type(values):
            code = "printf '%s\\t%s\\n' \\\n"
            for item, desc in values.items():
                code += '  %s %s \\\n' % (shell.quote(item), shell.quote(desc))
            code = code.rstrip(' \\\n')
        else:
            code = "printf '%s\\n' \\\n"
            for value in values:
                code += '  %s \\\n' % shell.quote(value)
            code = code.rstrip(' \\\n')

        func = ctxt.helpers.add_dynamic_func(ctxt, code)

        if duplicates:
            super().__init__(ctxt, ['__fish_complete_list', separator, func])
        else:
            complete_list_func = ctxt.helpers.use_function('list')
            super().__init__(ctxt, [complete_list_func, separator, func])


class FishCompletKeyValueList(FishCompletionCommand):
    '''Used for completing a list of key-value pairs.'''

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments

    def __init__(self, ctxt, trace, completer, pair_separator, value_separator, values):
        trace.append('key_value_list')
        args = []
        q = shell.quote

        for key, desc, complete in values:
            if not complete:
                func = 'false'
            elif complete[0] == 'none':
                func = 'true'
            else:
                obj = completer.complete_from_def(ctxt, trace, complete)
                func = obj.get_function()

            args.append('%s %s %s' % (q(key), q(desc or ''), q(func)))

        code = '%s %s %s \\\n%s' % (
            ctxt.helpers.use_function('key_value_list'),
            shell.quote(pair_separator),
            shell.quote(value_separator),
            indent(' \\\n'.join(args), 2)
        )

        func = ctxt.helpers.add_dynamic_func(ctxt, code)

        super().__init__(ctxt, [func])


class FishCompleteCommand(FishCompletionBase):
    '''Complete a command from $PATH.'''

    def __init__(self, ctxt, opts):
        super().__init__(ctxt)

        code = None
        path = None
        append = None
        prepend = None

        if opts:
            path = opts.get('path', None)
            append = opts.get('path_append', None)
            prepend = opts.get('path_prepend', None)

        def mkpath(path):
            return shell.join_quoted(path.split(':'))

        if path:
            code = 'set -lx PATH %s' % mkpath(path)
        elif append and prepend:
            code = 'set -lx PATH %s $PATH %s' % (mkpath(prepend), mkpath(append))
        elif append:
            code = 'set -lx -a PATH %s' % mkpath(append)
        elif prepend:
            code = 'set -lx PATH %s $PATH' % mkpath(prepend)

        if not code:
            self.code = "__fish_complete_command"
        else:
            self.code = f'{code}\n__fish_complete_command'

    def get_args(self):
        if '\n' in self.code:
            func = self.ctxt.helpers.add_dynamic_func(self.ctxt, self.code)
            return ['-f', '-a', '(%s)' % func]

        return ['-f', '-a', '(%s)' % self.code]

    def get_code(self):
        return self.code


class FishCompleteCombine(FishCompletionBase):
    '''Used for combining multiple complete commands.'''

    def __init__(self, ctxt, trace, completer, commands):
        super().__init__(ctxt)

        self.code = []
        trace.append('combine')

        for command in commands:
            obj = completer.complete_from_def(ctxt, trace, command)
            self.code.append(obj.get_code())

    def get_code(self):
        return '\n'.join(self.code)

    def get_args(self):
        code_is_singleline = not any('\n' in code for code in self.code)

        if code_is_singleline:
            return ['-f', '-a', '(%s)' % '; '.join(self.code)]

        code = '\n'.join(self.code)
        func = self.ctxt.helpers.add_dynamic_func(self.ctxt, code)
        return ['-f', '-a', '(%s)' % func]


class FishCompleteCommandArg(FishCompletionBase):
    '''Complete an argument of a command'''

    def __init__(self, ctxt):
        super().__init__(ctxt)

        query = ctxt.helpers.use_function('query', 'positionals_positions')
        types = get_defined_option_types(ctxt.option.parent.get_root_commandline())
        if types.short:
            ctxt.helpers.use_function('query', 'short_options')
        if types.long:
            ctxt.helpers.use_function('query', 'long_options')
        if types.old:
            ctxt.helpers.use_function('query', 'old_options')

        opts = get_query_option_strings(ctxt.option.parent, with_parent_options=True)
        opts = shell.quote(opts)
        command_pos = ctxt.option.get_positional_num() - 1

        r =  'set -l opts %s\n' % shell.quote(opts)
        r += 'set -l pos (%s "$opts" positional_pos %d)\n' % (query, command_pos)
        r += 'set -l cmdline (commandline -poc | string escape) (commandline -ct)\n'
        r += 'complete -C -- "$cmdline[$pos..]"'

        self.code = r

    def get_code(self):
        return self.code

    def get_args(self):
        func = self.ctxt.helpers.add_dynamic_func(self.ctxt, self.code)
        return ['-f', '-a', '(%s)' % func]

    def get_function(self):
        func = self.ctxt.helpers.add_dynamic_func(self.ctxt, self.code)
        return func


class FishCompleter(shell.ShellCompleter):
    '''Code generator for completing arguments in Fish.'''

    # pylint: disable=missing-function-docstring
    # pylint: disable=too-many-public-methods
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments

    def none(self, ctxt, _trace, *_):
        return FishCompleteNone(ctxt)

    def integer(self, ctxt, _trace, options=None):
        suffixes = []

        if options:
            if 'suffixes' in options:
                for suffix, description in options['suffixes'].items():
                    suffixes.append(f'{suffix}:{description}')

        if not suffixes:
            return FishCompleteNone(ctxt)

        func = ctxt.helpers.use_function('number')
        return FishCompletionCommand(ctxt, [func, *suffixes])

    def float(self, ctxt, trace, options=None):
        return self.integer(ctxt, trace, options)

    def choices(self, ctxt, _trace, choices):
        return FishCompleteChoices(ctxt, choices)

    def command(self, ctxt, _trace, opts=None):
        return FishCompleteCommand(ctxt, opts)

    def directory(self, ctxt, trace, opts=None):
        return FishCompleteDirectory(ctxt, trace, opts)

    def file(self, ctxt, _trace, opts=None):
        return FishCompleteFile(ctxt, opts)

    def mime_file(self, ctxt, _trace, pattern):
        func = ctxt.helpers.use_function('mime_file')
        return FishCompletionCommand(ctxt, [func, pattern])

    def group(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ["__fish_complete_groups"])

    def hostname(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ["__fish_print_hostnames"])

    def pid(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ["__fish_complete_pids"])

    def process(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ["__fish_complete_proc"])

    def range(self, ctxt, _trace, start, stop, step=1):
        if step == 1:
            return FishCompletionCommand(ctxt, ['seq', str(start), str(stop)])

        return FishCompletionCommand(ctxt, ['seq', str(start), str(step), str(stop)])

    def service(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ["__fish_systemctl_services"])

    def user(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ["__fish_complete_users"])

    def variable(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ["set", "-n"])

    def environment(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ["set", "-n", "-x"])

    def exec(self, ctxt, _trace, command):
        return FishCompletionRawCommand(ctxt, command)

    def exec_fast(self, ctxt, _trace, command):
        return FishCompletionRawCommand(ctxt, command)

    def exec_internal(self, ctxt, _trace, command):
        return FishCompletionRawCommand(ctxt, command)

    def value_list(self, ctxt, _trace, opts):
        return FishCompleteValueList(ctxt, opts)

    def key_value_list(self, ctxt, trace, pair_separator, value_separator, values):
        return FishCompletKeyValueList(ctxt, trace, self, pair_separator, value_separator, values)

    def combine(self, ctxt, trace, commands):
        return FishCompleteCombine(ctxt, trace, self, commands)

    def list(self, ctxt, trace, command, opts=None):
        separator = opts.get('separator', ',') if opts else ','
        duplicates = opts.get('duplicates', False) if opts else False
        trace.append('list')

        obj = self.complete_from_def(ctxt, trace, command)
        func = obj.get_function()

        list_func = ctxt.helpers.use_function('list')
        if not duplicates:
            return FishCompletionCommand(ctxt, [list_func, separator, func])

        return FishCompletionCommand(ctxt, [list_func, '-d', separator, func])

    def history(self, ctxt, _trace, pattern):
        func = ctxt.helpers.use_function('history')
        return FishCompletionCommand(ctxt, [func, pattern])

    def commandline_string(self, ctxt, _trace):
        func = ctxt.helpers.use_function('commandline_string')
        return FishCompletionCommand(ctxt, [func])

    def command_arg(self, ctxt, _trace):
        return FishCompleteCommandArg(ctxt)

    def date(self, ctxt, _trace, _format):
        return FishCompleteNone(ctxt)

    def date_format(self, ctxt, _trace):
        func = ctxt.helpers.use_function('date_format')
        return FishCompletionCommand(ctxt, [func])

    def uid(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ['__fish_complete_user_ids'])

    def gid(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ['__fish_complete_group_ids'])

    def filesystem_type(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ['__fish_print_filesystems'])

    def prefix(self, ctxt, trace, prefix, command):
        obj = self.complete_from_def(ctxt, trace, command)
        func = obj.get_function()

        prefix_func = ctxt.helpers.use_function('prefix')
        return FishCompletionCommand(ctxt, [prefix_func, prefix, func])

    def ip_address(self, ctxt, _trace, type_='all'):
        func = '__fish_print_addresses'

        if type_ == 'ipv6':
            cmd = f"{func} | string match -e ':'"
            return FishCompletionRawCommand(ctxt, cmd)

        if type_ == 'ipv4':
            cmd = f"{func} | string match -e '.'"
            return FishCompletionRawCommand(ctxt, cmd)

        return FishCompletionCommand(ctxt, [func])

    # =========================================================================
    # Bonus
    # =========================================================================

    def net_interface(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ['__fish_print_interfaces'])

    def mountpoint(self, ctxt, _trace):
        return FishCompletionCommand(ctxt, ['__fish_print_mounted'])

    def timezone(self, ctxt, _trace):
        func = ctxt.helpers.use_function('timezone_list')
        return FishCompletionCommand(ctxt, [func])

    def alsa_card(self, ctxt, _trace):
        func = ctxt.helpers.use_function('alsa_list_cards')
        return FishCompletionCommand(ctxt, [func])

    def alsa_device(self, ctxt, _trace):
        func = ctxt.helpers.use_function('alsa_list_devices')
        return FishCompletionCommand(ctxt, [func])



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains code for completing arguments in Zsh.'''

from . import shell
from .pattern import bash_glob_to_zsh_glob
from .str_utils import join_with_wrap, indent
from .zsh_utils import (
    escape_colon, escape_square_brackets, make_file_extension_pattern
)
from .type_utils import is_dict_type


CHOICES_INLINE_THRESHOLD = 80
# The `choices` command can in Zsh be expressed inline in an optspec, like:
#   (foo bar baz)
# or:
#   (foo:"Foo value" bar:"Bar value" baz:"Baz value)
#
# This variable defines how big this string can get before a function
# is used instead.


class ZshCompletionBase:
    '''Base class for Zsh completions.'''

    def get_action_string(self):
        '''Return an action string that can be used in an option spec.'''
        raise NotImplementedError

    def get_function(self):
        '''Return a function that can be used in e.g. `_sequence`.'''
        raise NotImplementedError


class ZshComplFunc(ZshCompletionBase):
    '''Complete using a function.'''

    def __init__(self, ctxt, args, needs_braces=False):
        self.ctxt = ctxt
        self.args = args
        self.needs_braces = needs_braces

    def get_action_string(self):
        if len(self.args) == 1:
            cmd = self.args[0]
        else:
            cmd = shell.join_quoted(self.args)

        if self.needs_braces:
            return shell.quote(f'{{{cmd}}}')

        return shell.quote(cmd)

    def get_function(self):
        if len(self.args) == 1:
            return self.args[0]

        code = shell.join_quoted(self.args)
        funcname = self.ctxt.helpers.add_dynamic_func(self.ctxt, code)
        return funcname


class ZshCompleteChoices(ZshCompletionBase):
    '''Complete from a set of words.'''

    def __init__(self, ctxt, trace, choices):
        self.ctxt = ctxt
        self.trace = trace
        self.choices = choices

    def _list_action_string(self):
        items = [str(item) for item in self.choices]
        quoted = [shell.quote(item) for item in items]
        action = shell.quote('(%s)' % ' '.join(quoted))
        if len(action) <= CHOICES_INLINE_THRESHOLD:
            return action

        return self._list_function()

    def _list_function(self):
        metavar = shell.quote(self.ctxt.option.metavar or '')
        quoted = [shell.quote(escape_colon(c)) for c in self.choices]
        line_length = self.ctxt.config.line_length - 2

        code = 'local items=(\n'
        code += indent(join_with_wrap(' ', '\n', line_length, quoted), 2)
        code += '\n)\n\n'
        code += f'_describe -- {metavar} items'

        funcname = self.ctxt.helpers.add_dynamic_func(self.ctxt, code)
        return funcname

    def _dict_action_string(self):
        def str0(s):
            return str(s) if s is not None else ''

        items = [str0(item) for item in self.choices.keys()]
        descriptions = [str0(value) for value in self.choices.values()]
        colon = any(':' in s for s in items + descriptions)
        quoted = []

        for item, desc in zip(items, descriptions):
            val = shell.quote(item)
            if desc:
                val += ':%s' % shell.quote(desc)
            quoted.append(val)

        action = shell.quote('((%s))' % ' '.join(quoted))
        if not colon and len(action) <= CHOICES_INLINE_THRESHOLD:
            return action

        return self._dict_function()

    def _dict_function(self):
        metavar = shell.quote(self.ctxt.option.metavar or '')

        code = 'local items=(\n'
        for item, desc in self.choices.items():
            item = shell.quote(escape_colon(str(item)))
            if desc:
                desc = shell.quote(str(desc))
                code += f'  {item}:{desc}\n'
            else:
                code += f'  {item}\n'
        code += ')\n\n'
        code += f'_describe -- {metavar} items'

        funcname = self.ctxt.helpers.add_dynamic_func(self.ctxt, code)
        return funcname

    def get_action_string(self):
        if is_dict_type(self.choices):
            return self._dict_action_string()

        return self._list_action_string()

    def get_function(self):
        if is_dict_type(self.choices):
            return self._dict_function()

        return self._list_function()


class ZshCompleteCommand(ZshCompletionBase):
    '''Complete a command from $PATH.'''

    def __init__(self, ctxt, opts):
        self.ctxt = ctxt
        self.code = None
        path = None
        append = None
        prepend = None

        if opts:
            path = opts.get('path', None)
            append = opts.get('path_append', None)
            prepend = opts.get('path_prepend', None)

        if path:
            self.code = 'local -x PATH=%s' % shell.quote(path)
        elif append and prepend:
            append = shell.quote(append)
            prepend = shell.quote(prepend)
            self.code = 'local -x PATH=%s:"$PATH":%s' % (prepend, append)
        elif append:
            self.code = 'local -x PATH="$PATH":%s' % shell.quote(append)
        elif prepend:
            self.code = 'local -x PATH=%s:"$PATH"' % shell.quote(prepend)

    def get_action_string(self):
        if not self.code:
            return '_command_names'

        code = f'{self.code}\n_command_names'
        funcname = self.ctxt.helpers.add_dynamic_func(self.ctxt, code)
        return funcname

    def get_function(self):
        if not self.code:
            return '_command_names'

        code = f'{self.code}\n_command_names'
        funcname = self.ctxt.helpers.add_dynamic_func(self.ctxt, code)
        return funcname


class ZshCompleteRange(ZshCompletionBase):
    '''Complete a range of integers.'''

    def __init__(self, ctxt, start, stop, step):
        self.ctxt = ctxt
        self.start = start
        self.stop = stop
        self.step = step

    def get_action_string(self):
        if self.step == 1:
            return f"'({{{self.start}..{self.stop}}})'"

        return f"'({{{self.start}..{self.stop}..{self.step}}})'"

    def get_function(self):
        if self.step == 1:
            code = f"command seq {self.start} {self.stop}"
        else:
            code = f"command seq {self.start} {self.step} {self.stop}"

        code = f'compadd -- $({code})'

        funcname = self.ctxt.helpers.add_dynamic_func(self.ctxt, code)
        return funcname


class ZshKeyValueList(ZshComplFunc):
    '''Used for completing a list of key-value pairs.'''

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments

    def __init__(self, ctxt, trace, completer, pair_separator, value_separator, values):
        trace.append('key_value_list')
        spec = []

        for key, desc, complete in values:
            key = escape_colon(key)

            if desc:
                desc = shell.quote('[%s]' % escape_square_brackets(desc))
            else:
                desc = ''

            if not complete:
                spec.append(f'{key}{desc}')
            elif complete[0] == 'none':
                spec.append(f'{key}{desc}:::')
            else:
                compl_obj = completer.complete_from_def(ctxt, trace, complete)
                action = compl_obj.get_action_string()
                spec.append(f'{key}{desc}:::{action}')

        code = '_values -s %s -S %s %s \\\n%s' % (
            shell.quote(pair_separator),
            shell.quote(value_separator),
            shell.quote(ctxt.option.metavar or ''),
            indent(' \\\n'.join(spec), 2)
        )

        func = ctxt.helpers.add_dynamic_func(ctxt, code)

        super().__init__(ctxt, [func], needs_braces=True)


class ZshCompleteCombine(ZshCompletionBase):
    '''Combine multiple completers into one.'''

    def __init__(self, ctxt, trace, completer, commands):
        trace.append('combine')

        completions = []
        for command_args in commands:
            compl_obj = completer.complete_from_def(ctxt, trace, command_args)
            action_string = compl_obj.get_action_string()
            completions.append(f'::{action_string}')

        code = '_alternative \\\n'
        code += indent(' \\\n'.join(completions), 2)

        self.func = ctxt.helpers.add_dynamic_func(ctxt, code)

    def get_function(self):
        return self.func

    def get_action_string(self):
        return shell.quote('{%s}' % self.func)


class ZshCompleter(shell.ShellCompleter):
    '''Code generator for completing arguments in Zsh.'''

    # pylint: disable=too-many-public-methods
    # pylint: disable=missing-function-docstring
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments

    def none(self, ctxt, _trace, *_):
        return ZshComplFunc(ctxt, [' '])

    def integer(self, ctxt, _trace, options=None):
        args = []
        suffixes = []
        help_ = ctxt.option.help or ctxt.option.metavar or ''

        if options:
            if 'min' in options:
                args += ['-l', str(options['min'])]
            if 'max' in options:
                args += ['-m', str(options['max'])]
            if 'suffixes' in options:
                for suffix, description in options['suffixes'].items():
                    suffixes.append(f'{suffix}:{description}')
            if 'help' in options:
                help_ = options['help']

        return ZshComplFunc(ctxt, ['_numbers', *args, help_, *suffixes])

    def float(self, ctxt, _trace, options=None):
        args = ['-f']
        suffixes = []
        help_ = ctxt.option.help or ctxt.option.metavar or ''

        if options:
            if 'min' in options:
                args += ['-l', str(options['min'])]
            if 'max' in options:
                args += ['-m', str(options['max'])]
            if 'suffixes' in options:
                for suffix, description in options['suffixes'].items():
                    suffixes.append(f'{suffix}:{description}')
            if 'help' in options:
                help_ = options['help']

        return ZshComplFunc(ctxt, ['_numbers', *args, help_, *suffixes])

    def choices(self, ctxt, trace, choices):
        return ZshCompleteChoices(ctxt, trace, choices)

    def command(self, ctxt, _trace, opts=None):
        return ZshCompleteCommand(ctxt, opts)

    def directory(self, ctxt, _trace, opts=None):
        directory = None if opts is None else opts.get('directory', None)

        if not directory:
            return ZshComplFunc(ctxt, ['_directories'])

        if directory.startswith('/'):
            return ZshComplFunc(ctxt, ['_directories', '-W', directory])

        func = ctxt.helpers.use_function('path_files_relative')
        return ZshComplFunc(ctxt, [func, directory, '-/'], needs_braces=True)

    def file(self, ctxt, _trace, opts=None):
        fuzzy = False
        directory = None
        extensions = None
        ignore_globs = None

        if opts:
            fuzzy = opts.get('fuzzy', False)
            directory = opts.get('directory', None)
            extensions = opts.get('extensions', None)
            ignore_globs = opts.get('ignore_globs', None)

        args = []

        if extensions:
            args.extend(['-g', make_file_extension_pattern(extensions, fuzzy)])

        if ignore_globs:
            patterns = map(bash_glob_to_zsh_glob, ignore_globs)
            args.extend(['-F', '(%s)' % ' '.join(patterns)])

        if not directory:
            return ZshComplFunc(ctxt, ['_files'] + args)

        if directory.startswith('/'):
            return ZshComplFunc(ctxt, ['_files', '-W', directory, *args])

        func = ctxt.helpers.use_function('path_files_relative')
        return ZshComplFunc(ctxt, [func, directory, *args], needs_braces=True)

    def mime_file(self, ctxt, _trace, pattern):
        func = ctxt.helpers.use_function('mime_file')
        return ZshComplFunc(ctxt, [func, pattern], needs_braces=True)

    def group(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_groups'])

    def hostname(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_hosts'])

    def pid(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_pids'])

    def process(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_process_names', '-a'])

    def range(self, ctxt, _trace, start, stop, step=1):
        return ZshCompleteRange(ctxt, start, stop, step)

    def user(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_users'])

    def variable(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_vars'])

    def environment(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_parameters', '-g', '*export*'])

    def exec(self, ctxt, _trace, command):
        funcname = ctxt.helpers.use_function('exec')
        return ZshComplFunc(ctxt, [funcname, command], needs_braces=True)

    def exec_fast(self, ctxt, _trace, command):
        funcname = ctxt.helpers.use_function('exec')
        return ZshComplFunc(ctxt, [funcname, command], needs_braces=True)

    def exec_internal(self, ctxt, _trace, command):
        return ZshComplFunc(ctxt, [command], needs_braces=True)

    def value_list(self, ctxt, _trace, opts):
        desc = ctxt.option.metavar or ''
        values = opts['values']
        separator = opts.get('separator', ',')
        duplicates = opts.get('duplicates', False)

        if is_dict_type(values):
            esc = escape_square_brackets
            values = ['%s[%s]' % (esc(item), esc(desc)) for item, desc in values.items()]
        else:
            values = [escape_square_brackets(i) for i in values]

        if not duplicates:
            return ZshComplFunc(ctxt, ['_values', '-s', separator, desc] + values)

        values_func = ZshComplFunc(ctxt, ['_values', desc] + values).get_function()

        if separator == ',':
            return ZshComplFunc(ctxt, ['_sequence', '-d', values_func])

        return ZshComplFunc(ctxt, ['_sequence', '-s', separator, '-d', values_func])

    def key_value_list(self, ctxt, trace, pair_separator, value_separator, values):
        return ZshKeyValueList(ctxt, trace, self, pair_separator, value_separator, values)

    def combine(self, ctxt, trace, commands):
        return ZshCompleteCombine(ctxt, trace, self, commands)

    def list(self, ctxt, trace, command, opts=None):
        separator = opts.get('separator', ',') if opts else ','
        duplicates = opts.get('duplicates', False) if opts else False
        trace.append('list')

        obj = self.complete_from_def(ctxt, trace, command)
        func = obj.get_function()

        args = []

        if separator != ',':
            args.extend(['-s', separator])

        if duplicates:
            args.append('-d')

        return ZshComplFunc(ctxt, ['_sequence', *args, func])

    def history(self, ctxt, _trace, pattern):
        func = ctxt.helpers.use_function('history')
        return ZshComplFunc(ctxt, [func, pattern], needs_braces=True)

    def commandline_string(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_cmdstring'])

    def command_arg(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_normal'])

    def date(self, ctxt, _trace, format_):
        return ZshComplFunc(ctxt, ['_dates', '-f', format_])

    def date_format(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_date_formats'])

    def uid(self, ctxt, _trace):
        func = ctxt.helpers.use_function('uid_list')
        return ZshComplFunc(ctxt, [func], needs_braces=True)

    def gid(self, ctxt, _trace):
        func = ctxt.helpers.use_function('gid_list')
        return ZshComplFunc(ctxt, [func], needs_braces=True)

    def filesystem_type(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_file_systems'])

    def signal(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_signals'])

    def prefix(self, ctxt, trace, prefix, command):
        prefix_func = ctxt.helpers.use_function('prefix')
        obj = self.complete_from_def(ctxt, trace, command)
        func = obj.get_function()
        return ZshComplFunc(ctxt, [prefix_func, prefix, func], needs_braces=True)

    def ip_address(self, ctxt, _trace, type_='all'):
        if type_ == 'ipv4':
            return ZshComplFunc(ctxt, ['_bind_addresses', '-4'])

        if type_ == 'ipv6':
            return ZshComplFunc(ctxt, ['_bind_addresses', '-6'])

        return ZshComplFunc(ctxt, ['_bind_addresses'])

    # =========================================================================
    # Bonus
    # =========================================================================

    def mountpoint(self, ctxt, _trace):
        func = ctxt.helpers.use_function('mountpoint')
        return ZshComplFunc(ctxt, [func], needs_braces=True)

    def net_interface(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_net_interfaces'])

    def timezone(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_time_zone'])

    def locale(self, ctxt, _trace):
        return ZshComplFunc(ctxt, ['_locales'])

    def charset(self, ctxt, _trace):
        func = ctxt.helpers.use_function('charset_list')
        return ZshComplFunc(ctxt, [func], needs_braces=True)

    def alsa_card(self, ctxt, _trace):
        func = ctxt.helpers.use_function('alsa_complete_cards')
        return ZshComplFunc(ctxt, [func], needs_braces=True)

    def alsa_device(self, ctxt, _trace):
        func = ctxt.helpers.use_function('alsa_complete_devices')
        return ZshComplFunc(ctxt, [func], needs_braces=True)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Class for main command line application.'''

import os
import sys
import shlex
import argparse

from .errors import CrazyError
from . import bash, fish, zsh
from . import argparse_source, json_source, yaml_source
from . import argparse_mod  # .complete()
from . import help_converter
from . import utils
from . import config
from . import paths


# The import of `argparse_mod` only modifies the classes provided by the
# argparse module. We use the dummy function to silence warnings about the
# unused import.
argparse_mod.dummy()


def boolean(string):
    '''Convert string to bool, else raise ValueError.'''

    try:
        return {'true': True, 'false': False}[string.lower()]
    except KeyError as e:
        raise ValueError(f"Not a bool: {string}") from e


def version(string):
    '''Parse a version.'''

    return tuple(map(int, string.split('.')))


def feature_list(string):
    '''Convert a comma separated string of features to list.'''

    features = string.split(',')

    for feature in features:
        if feature not in ('hidden', 'final', 'groups', 'repeatable', 'when'):
            raise ValueError(f"Invalid feature: {feature}")

    return features

_URL = "https://github.com/crazy-complete/crazy-complete"
_EPILOG = f"For complete documentation, see {_URL}"

p = argparse.ArgumentParser(
    'crazy-complete',
    description='Generate shell auto completion files for all major shells',
    epilog=_EPILOG,
    exit_on_error=False)

p.add_argument(
    'shell', choices=('bash', 'fish', 'zsh', 'json', 'yaml'),
    help='Specify the shell type for the completion script')

p.add_argument(
    'definition_file',
    help='The file containing the command line definitions'
).complete('file')

p.add_argument(
    '--version', action='version', version='%(prog)s 0.3.8',
    help='Show program version')

p.add_argument(
    '--manual', metavar='TOPIC', nargs='?',
    help='Show manual for topic')

p.add_argument(
    '--parser-variable', default=None,
    help='Specify the variable name of the ArgumentParser object (for --input-type=python)')

p.add_argument(
    '--input-type', choices=('yaml', 'json', 'python', 'help', 'auto'), default='auto',
    help='Specify input file type')

p.add_argument(
    '--abbreviate-commands', metavar='BOOL', default=False, type=boolean,
    help='Sets whether commands can be abbreviated'
).complete('choices', ('True', 'False'))

p.add_argument(
    '--abbreviate-options', metavar='BOOL', default=False, type=boolean,
    help='Sets whether options can be abbreviated'
).complete('choices', ('True', 'False'))

p.add_argument(
    '--repeatable-options', metavar='BOOL', default=False, type=boolean,
    help='Sets whether options are suggested multiple times during completion'
).complete('choices', ('True', 'False'))

p.add_argument(
    '--inherit-options', metavar='BOOL', default=False, type=boolean,
    help='Sets whether parent options are visible to subcommands'
).complete('choices', ('True', 'False'))

p.add_argument(
    '--option-stacking', metavar='BOOL', default=True, type=boolean,
    help='Sets whether short option stacking is allowed'
).complete('choices', ('True', 'False'))

p.add_argument(
    '--long-option-argument-separator', default='both',
    choices=('space', 'equals', 'both'),
    dest='long_opt_arg_sep',
    help='Sets which separators are used for delimiting a long option from its argument'
).complete('choices', ('space', 'equals', 'both'))

p.add_argument(
    '--disable', metavar='FEATURES', default=[], type=feature_list,
    help='Disable features (hidden,final,groups,repeatable,when)'
).complete('value_list', {'values': {
    'hidden':     'Disable hidden options',
    'final':      'Disable final options',
    'groups':     'Disable option groups',
    'repeatable': 'Disable check for repeatable options',
    'when':       'Disable conditional options and positionals',
}})

p.add_argument(
    '--vim-modeline', metavar='BOOL', default=True, type=boolean,
    help='Sets whether a vim modeline comment shall be appended to the generated code'
).complete('choices', ('True', 'False'))

p.add_argument(
    '--bash-completions-version', metavar='VERSION', default=(2,), type=version,
    help='Generate code for a specific bash-completions version'
)

p.add_argument(
    '--zsh-compdef', metavar='BOOL', default=True, type=boolean,
    help='Sets whether #compdef is used in zsh scripts'
).complete('choices', ('True', 'False'))

p.add_argument(
    '--fish-fast', metavar='BOOL', default=False, type=boolean,
    help='Use faster commandline parsing at the cost of correctness'
).complete('choices', ('True', 'False'))

p.add_argument(
    '--fish-inline-conditions', metavar='BOOL', default=False, type=boolean,
    help="Don't store conditions in a variable"
).complete('choices', ('True', 'False'))

p.add_argument(
    '--include-file', metavar='FILE', action='append',
    help='Include file in output'
).complete('file')

p.add_argument(
    '--comment', metavar='COMMENT', action='append',
    help='Add a comment to output'
)

p.add_argument(
    '--debug', action='store_true', default=False,
    help='Enable debug mode')

p.add_argument(
    '--keep-comments', action='store_true', default=False,
    help='Keep comments in generated output')

p.add_argument(
    '--max-line-length', default=80, type=int,
    help='Set the maximum line length of the generated output'
)

p.add_argument(
    '--function-prefix', metavar='PREFIX', default='_$PROG',
    help='Set the prefix used for generated functions')

grp = p.add_mutually_exclusive_group()

grp.add_argument(
    '-o', '--output', metavar='FILE', default=None, dest='output_file',
    help='Write output to destination file [default: stdout]'
).complete('file')

grp.add_argument(
    '-i', '--install-system-wide', default=False, action='store_true',
    help='Write output to the system-wide completions dir of shell')

grp.add_argument(
    '-u', '--uninstall-system-wide', default=False, action='store_true',
    help='Uninstall the system-wide completion file for program')

# We use a unique object name for avoiding name clashes when
# importing/executing the foreign python script
_crazy_complete_argument_parser = p
del p


def write_string_to_file(string, file):
    '''Writes string to file if file is given.'''

    if file is not None:
        with open(file, 'w', encoding='utf-8') as fh:
            fh.write(string)
    else:
        print(string)


def load_definition_file(opts):
    '''Load a definition file as specified in `opts`.'''

    if opts.input_type == 'auto':
        basename = os.path.basename(opts.definition_file)
        extension = os.path.splitext(basename)[1].lower().strip('.')

        if extension == 'json':
            return json_source.load_from_file(opts.definition_file)

        if extension in ('yaml', 'yml'):
            return yaml_source.load_from_file(opts.definition_file)

        if extension == 'py':
            msg = 'Reading Python files must be enabled by --input-type=python'
            raise CrazyError(msg)

        if extension == '':
            msg = ('File has no extension. '
                   'Please supply --input-type=json|yaml|help|python')
            raise CrazyError(msg)

        msg = (f'Unknown file extension `{extension}`. '
               'Please supply --input-type=json|yaml|help|python')
        raise CrazyError(msg)

    if opts.input_type == 'json':
        return json_source.load_from_file(opts.definition_file)

    if opts.input_type == 'yaml':
        return yaml_source.load_from_file(opts.definition_file)

    if opts.input_type == 'python':
        return argparse_source.load_from_file(
            opts.definition_file,
            opts.parser_variable,
            parser_blacklist=[_crazy_complete_argument_parser])

    raise AssertionError("Should not be reached")


def _get_config_from_options(opts):
    conf = config.Config()
    conf.set_debug(opts.debug)
    conf.set_function_prefix(opts.function_prefix)
    conf.set_abbreviate_commands(opts.abbreviate_commands)
    conf.set_abbreviate_options(opts.abbreviate_options)
    conf.set_repeatable_options(opts.repeatable_options)
    conf.set_inherit_options(opts.inherit_options)
    conf.set_option_stacking(opts.option_stacking)
    conf.set_long_option_argument_separator(opts.long_opt_arg_sep)
    conf.set_vim_modeline(opts.vim_modeline)
    conf.set_bash_completions_version(opts.bash_completions_version)
    conf.set_zsh_compdef(opts.zsh_compdef)
    conf.set_fish_fast(opts.fish_fast)
    conf.set_fish_inline_conditions(opts.fish_inline_conditions)
    conf.include_many_files(opts.include_file or [])
    conf.set_keep_comments(opts.keep_comments)
    conf.add_comments(opts.comment or [])
    conf.set_line_length(opts.max_line_length)

    for feature in opts.disable:
        if feature == 'hidden':
            conf.disable_hidden(True)
        elif feature == 'final':
            conf.disable_final(True)
        elif feature == 'groups':
            conf.disable_groups(True)
        elif feature == 'repeatable':
            conf.disable_repeatable(True)
        elif feature == 'when':
            conf.disable_when(True)

    return conf


def try_print_manual(args):
    '''Print manual.'''

    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--manual', nargs='?', const='all')
    opts, _ = parser.parse_known_args(args)

    if not opts.manual:
        return

    from . import manual
    manual.print_help_topic(opts.manual, os.isatty(sys.stdout.fileno()))
    sys.exit(0)


def try_batch(args):
    '''Do batch processing.'''

    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--batch')
    opts, _ = parser.parse_known_args(args)

    if not opts.batch:
        return

    with open(opts.batch, 'r', encoding='utf-8') as fh:
        content = fh.read()

    for line in content.split('\n'):
        line = line.strip()

        if not line or line.startswith('#'):
            continue

        utils.print_err('Running:', sys.argv[0], line)

        app = Application()
        app.parse_args(shlex.split(line))
        app.run()

    sys.exit(0)


def generate(opts):
    '''Generate output file as specified in `opts`.'''

    if opts.input_type == 'help':
        if opts.shell != 'yaml':
            raise CrazyError('The `help` input-type currently only supports YAML generation')
        output = help_converter.from_file_to_yaml(opts.definition_file)
        write_string_to_file(output, opts.output_file)
        return

    cmdline = load_definition_file(opts)

    if opts.shell == 'json':
        output = json_source.commandline_to_json(cmdline)
        write_string_to_file(output, opts.output_file)
        return

    if opts.shell == 'yaml':
        output = yaml_source.commandline_to_yaml(cmdline)
        write_string_to_file(output, opts.output_file)
        return

    conf = _get_config_from_options(opts)

    if opts.shell == 'bash':
        output = bash.generate_completion(cmdline, conf)
    elif opts.shell == 'fish':
        output = fish.generate_completion(cmdline, conf)
    elif opts.shell == 'zsh':
        output = zsh.generate_completion(cmdline, conf)

    if opts.install_system_wide or opts.uninstall_system_wide:
        file = {
            'bash':  paths.get_bash_completion_file,
            'fish':  paths.get_fish_completion_file,
            'zsh':   paths.get_zsh_completion_file,
        }[opts.shell](cmdline.prog)

        if opts.install_system_wide:
            utils.print_err(f'Installing to {file}')
            write_string_to_file(output, file)
        else:
            utils.print_err(f'Removing {file}')
            os.remove(file)
    else:
        write_string_to_file(output, opts.output_file)


class Application:
    '''Class for main command line application.'''

    def __init__(self):
        self.options = None

    def parse_args(self, args):
        '''Parse command line arguments.

        Raises:
            - argparse.ArgumentError
        '''
        try_print_manual(args)
        try_batch(args)
        self.options = _crazy_complete_argument_parser.parse_args(args)

    def run(self):
        '''Run the crazy-complete program.

        Raises:
            - CrazyError
            - FileNotFoundError
            - yaml.scanner.ScannerError
            - yaml.parser.ParserError
            - json.decoder.JSONDecodeError
        '''
        generate(self.options)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

"""
This module provides functions for creating CommandLine objects from JSON
and vice versa.
"""

import json

from . import dictionary_source


def load_from_file(file):
    '''Load a JSON file and turn it into a cli.CommandLine object.'''

    with open(file, 'r', encoding='utf-8') as fh:
        dictionaries = json.load(fh)
        return dictionary_source.dictionaries_to_commandline(dictionaries)


def commandline_to_json(commandline):
    '''Convert a cli.CommandLine object into JSON.'''

    dictionaries = dictionary_source.commandline_to_dictionaries(commandline)
    json_string = json.dumps(dictionaries, indent=None)
    return json_string



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Module for option completion in Bash.'''

from . import cli
from . import algo
from . import utils
from . import bash_when
from . import bash_utils
from . import bash_patterns
from .str_utils import indent


class MasterCompletionFunctionBase:
    '''Base class for generating master completion functions.'''

    def __init__(self, options, abbreviations, generator):
        self.options = options
        self.abbreviations = abbreviations
        self.generator = generator
        self.complete = generator.complete_option

    def get(self, funcname):
        '''Get the function code.'''
        raise NotImplementedError

    def get_all_option_strings(self, option):
        '''Return all possible option strings for option.'''
        opts = option.get_short_option_strings()

        opts += self.abbreviations.get_many_abbreviations(
                option.get_long_option_strings())

        opts += self.abbreviations.get_many_abbreviations(
                option.get_old_option_strings())

        return opts


class MasterCompletionFunction(MasterCompletionFunctionBase):
    '''Class for generating a master completion function.'''

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods

    def __init__(self, options, abbreviations, generator):
        super().__init__(options, abbreviations, generator)
        self.optionals = False
        self.code = []

        optional_arg = list(filter(cli.Option.has_optional_arg, options))
        required_arg = list(filter(cli.Option.has_required_arg, options))

        self._add_options(required_arg)

        if optional_arg:
            self.optionals = True
            self.code.append('[[ "$mode" == WITH_OPTIONAL ]] || return 1')
            self._add_options(optional_arg)

    def _add_options(self, options):
        with_when, without_when = algo.partition(options, lambda o: o.when)
        self._add_options_with_when(with_when)
        self._add_options_without_when(without_when)

    def _add_options_without_when(self, options_):
        options_group_by_complete = algo.group_by(options_, self.complete)

        if options_group_by_complete:
            r = 'case "$opt" in\n'
            for complete, options in options_group_by_complete.items():
                opts = algo.flatten(map(self.get_all_option_strings, options))
                r += '  %s)\n' % bash_patterns.make_pattern(opts)
                if complete:
                    r += '%s\n' % indent(complete, 4)
                r += '    return 0;;\n'
            r += 'esac'
            self.code.append(r)

    def _add_options_with_when(self, options):
        for option in options:
            opts = self.get_all_option_strings(option)
            completion_code = self.complete(option)
            cond = bash_when.generate_when_conditions(
                self.generator.commandline,
                self.generator.variable_manager,
                option.when)

            r  = 'case "$opt" in %s)\n' % bash_patterns.make_pattern(opts)
            r += '  if %s; then\n' % cond
            if completion_code:
                r += '%s\n' % indent(completion_code, 4)
            r += '    return 0\n'
            r += '  fi;;\n'
            r += 'esac'
            self.code.append(r)

    def get(self, funcname):
        '''Get the function code.'''

        if self.code:
            r  = '%s() {\n' % funcname
            if self.optionals:
                r += '  local opt="$1" cur="$2" mode="$3"\n\n'
            else:
                r += '  local opt="$1" cur="$2"\n\n'
            r += '%s\n\n' % indent('\n\n'.join(self.code), 2)
            r += '  return 1\n'
            r += '}'
            return r

        return None


class MasterCompletionFunctionNoWhen(MasterCompletionFunctionBase):
    '''Class for generating a master completion function.'''

    def __init__(self, options, abbreviations, generator):
        super().__init__(options, abbreviations, generator)
        self.optionals = False
        self.code = []

        optional_arg = list(filter(cli.Option.has_optional_arg, options))
        required_arg = list(filter(cli.Option.has_required_arg, options))

        self._add_options(required_arg)

        if optional_arg:
            self.optionals = True
            r =  '(( ! ret )) && return 0\n'
            r += '[[ "$mode" == WITH_OPTIONAL ]] || return 1\n'
            r += 'ret=0'
            self.code.append(r)
            self._add_options(optional_arg)

    @staticmethod
    def accept(options):
        '''Return True if none of the options have the when attribute set.'''
        return not any(option.when for option in options)

    def _add_options(self, options_):
        options_group_by_complete = algo.group_by(options_, self.complete)

        if options_group_by_complete:
            r = 'case "$opt" in\n'
            for complete, options in options_group_by_complete.items():
                opts = algo.flatten(map(self.get_all_option_strings, options))
                if complete:
                    r += '  %s)\n' % bash_patterns.make_pattern(opts)
                    r += '%s;;\n' % indent(complete, 4)
                else:
                    r += '  %s);;\n' % bash_patterns.make_pattern(opts)
            r += '  *) ret=1;;\n'
            r += 'esac'
            self.code.append(r)

    def get(self, funcname):
        '''Get the function code.'''

        if self.code:
            r  = '%s() {\n' % funcname
            if self.optionals:
                r += '  local opt="$1" cur="$2" mode="$3" ret=0\n\n'
            else:
                r += '  local opt="$1" cur="$2" ret=0\n\n'
            r += '%s\n\n' % indent('\n\n'.join(self.code), 2)
            r += '  return $ret\n'
            r += '}'
            return r

        return None


def _get_master_completion_obj(options, abbreviations, generator):
    if MasterCompletionFunctionNoWhen.accept(options):
        klass = MasterCompletionFunctionNoWhen
    else:
        klass = MasterCompletionFunction

    return klass(options, abbreviations, generator)


class _Info:

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods

    def __init__(self, options, abbreviations, commandline, ctxt):
        self.commandline    = commandline
        self.ctxt           = ctxt
        self.short_required = False  # Short with required argument
        self.short_optional = False  # Short with optional argument
        self.long_required  = False  # Long with required argument
        self.long_optional  = False  # Long with optional argument
        self.old_required   = False  # Old-Style with required argument
        self.old_optional   = False  # Old-Style with optional argument

        self._collect_options_info(options)

        all_options = self.commandline.get_options(
            with_parent_options=self.commandline.inherit_options)

        old_option_strings = algo.flatten([o.get_old_option_strings() for o in all_options])
        old_option_strings = abbreviations.get_many_abbreviations(old_option_strings)
        self.old_option_strings = bash_utils.CasePatterns.for_old_without_arg(old_option_strings)

        self.short_no_args = ''
        self.short_required_args = ''
        self.short_optional_args = ''

        for option in all_options:
            short_opts = ''.join(o.lstrip('-') for o in option.get_short_option_strings())

            if option.has_required_arg():
                self.short_required_args += short_opts
            elif option.has_optional_arg():
                self.short_optional_args += short_opts
            elif option.complete is None:
                self.short_no_args += short_opts

        if self.short_no_args:
            self.short_no_args_pattern = '*([%s])' % self.short_no_args
        else:
            self.short_no_args_pattern = ''

    def _collect_options_info(self, options):
        for option in options:
            if option.get_long_option_strings():
                if option.has_optional_arg():
                    self.long_optional = True
                elif option.has_required_arg():
                    self.long_required = True

            if option.get_old_option_strings():
                if option.has_optional_arg():
                    self.old_optional = True
                elif option.has_required_arg():
                    self.old_required = True

            if option.get_short_option_strings():
                if option.has_optional_arg():
                    self.short_optional = True
                elif option.has_required_arg():
                    self.short_required = True


def _get_prev_completion(info):
    r = 'case "$prev" in\n'

    if info.long_required:
        r += '  --*) __complete_option "$prev" "$cur" WITHOUT_OPTIONAL && return 0;;\n'
    else:
        r += '  --*);;\n'

    if info.old_required and info.short_required:
        r += '  -*)  __complete_option "$prev" "$cur" WITHOUT_OPTIONAL && return 0;&\n'
        r += '  -%s[%s])\n' % (info.short_no_args_pattern, info.short_required_args)
        r += '       __complete_option "-${prev: -1}" "$cur" WITHOUT_OPTIONAL && return 0;;\n'
    elif info.old_required:
        r += '  -*)  __complete_option "$prev" "$cur" WITHOUT_OPTIONAL && return 0;;\n'
    elif info.short_required:
        r += '  -%s[%s])\n' % (info.short_no_args_pattern, info.short_required_args)
        r += '       __complete_option "-${prev: -1}" "$cur" WITHOUT_OPTIONAL && return 0;;\n'

    r += 'esac'
    return r


def _get_cur_completion(info):
    short_no_args_pattern = info.short_no_args_pattern
    short_required_args = info.short_required_args
    short_optional_args = info.short_optional_args

    r = 'case "$cur" in\n'

    if info.long_required or info.long_optional:
        r += '  --*=*)\n'
        r += '    __complete_option "${cur%%=*}" "${cur#*=}" WITH_OPTIONAL && return 0;;\n'
    else:
        r += '  --*=*);;\n'

    r += '  --*);;\n'

    if (info.short_required or info.short_optional) and info.old_option_strings:
        r += '  %s);;\n' % info.old_option_strings

    if info.old_required or info.old_optional:
        r += '  -*=*)\n'
        r += '    __complete_option "${cur%%=*}" "${cur#*=}" WITH_OPTIONAL && return 0;&\n'

    if info.short_required or info.short_optional:
        r += '  -%s[%s%s]*)\n' % (short_no_args_pattern, short_required_args, short_optional_args)
        r += '    local i\n'
        r += '    for ((i=2; i <= ${#cur}; ++i)); do\n'
        r += '      local pre="${cur:0:$i}" value="${cur:$i}"\n'
        r += '      __complete_option "-${pre: -1}" "$value" WITH_OPTIONAL && {\n'
        r += '        %s "$pre"\n' % info.ctxt.helpers.use_function('prefix_compreply')
        r += '        return 0\n'
        r += '      }\n'
        r += '    done;;\n'

    # Optimization: Strip unused case patterns at the end
    if r.endswith('  -*=*);;\n'):
        r = r.replace('  -*=*);;\n', '')

    if r.endswith('  --*);;\n'):
        r = r.replace('  --*);;\n', '')

    r += 'esac'
    return r


def _get_dispatcher(info):
    r = []

    required = info.short_required or info.long_required or info.old_required
    optional = info.short_optional or info.long_optional or info.old_optional

    if required:
        r.append(_get_prev_completion(info))

    if required or optional:
        r.append(_get_cur_completion(info))

    return '\n\n'.join(r)


def generate_option_completion(self):
    '''Generate code for completing options.'''

    options = self.commandline.get_options(only_with_arguments=True)
    abbreviations = utils.get_option_abbreviator(self.commandline)

    complete_function = _get_master_completion_obj(options, abbreviations, self)
    complete_function_code = complete_function.get('__complete_option')

    if not complete_function_code:
        return None

    info = _Info(options, abbreviations, self.commandline, self.ctxt)
    dispatcher_code = _get_dispatcher(info)

    if not complete_function.optionals:
        dispatcher_code = dispatcher_code.replace('WITH_OPTIONAL ', '')
        dispatcher_code = dispatcher_code.replace('WITHOUT_OPTIONAL ', '')

    return f'{complete_function_code}\n\n{dispatcher_code}'



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Version-dependent Bash code.'''


# pylint: disable=missing-function-docstring


def filedir(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_compgen_filedir'

    return '_filedir'


def init_completion(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_initialize'

    return '_init_completion'


def completion_loader(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_complete_load'

    return '_completion_loader'


def pids(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_compgen_pids'

    return '_pids'


def pnames(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_compgen_pnames'

    return '_pnames'


def uids(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_compgen_uids'

    return '_uids'


def gids(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_compgen_gids'

    return '_gids'


def shells(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_compgen -a shells'

    return '_shells'


def signals(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_compgen_signals'

    return '_signals'


def fstypes(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_compgen -a fstypes'

    return '_fstypes'


def command_offset(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_command_offset'

    return '_command_offset'


def available_interfaces(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_compgen_available_interfaces'

    return '_available_interfaces'


def ip_addresses(ctxt):
    if ctxt.config.bash_completions_version >= (2, 12):
        return '_comp_compgen_ip_addresses'

    return '_ip_addresses'



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Classes for including functions in the generation process.'''

import re

from . import cli
from . import shell
from . import preprocessor
from .str_utils import (indent, strip_comments, strip_double_empty_lines)


# pylint: disable=too-few-public-methods


def replace_function_name(search, replace, s):
    '''Safely replace a function name in code'''

    pattern = f'\\b{search}\\b'
    return re.sub(pattern, replace, s)


class FunctionBase:
    '''Base class for functions.'''

    def __init__(self, funcname, code, dependencies=None):
        self.funcname = funcname
        self.code = code.strip()
        self.dependencies = dependencies if dependencies else []

    def get_code(self):
        '''Return the whole function.'''
        raise NotImplementedError


class ShellFunction(FunctionBase):
    '''Class for generating Bash and Zsh functions.'''

    def get_code(self, funcname=None, defines=frozenset()):
        if funcname is None:
            funcname = self.funcname

        code = self.code.replace('%FUNCNAME%', funcname)
        code = preprocessor.preprocess(code, defines)
        code = strip_double_empty_lines(code)

        r  = '%s() {\n' % funcname
        r += '%s\n'     % indent(code, 2).rstrip()
        r += '}'
        return r


class FishFunction(FunctionBase):
    '''Class for generating Fish functions.'''

    def get_code(self, funcname=None, defines=frozenset()):
        if funcname is None:
            funcname = self.funcname

        code = self.code.replace('%FUNCNAME%', funcname)
        code = preprocessor.preprocess(code, defines)
        code = strip_double_empty_lines(code)

        r  = 'function %s\n' % funcname
        r += '%s\n'          % indent(code, 2).rstrip()
        r += 'end'
        return r


def make_completion_funcname_for_context(ctxt, prefix):
    '''TODO.'''

    commandlines = ctxt.commandline.get_parents(include_self=True)
    prognames = [p.prog for p in commandlines]
    prognames[0] = prefix

    funcname = shell.make_identifier('_'.join(prognames))

    if isinstance(ctxt.option, cli.Option):
        return '%s__%s' % (funcname, ctxt.option.option_strings[0])
    if isinstance(ctxt.option, cli.Positional):
        if ctxt.option.metavar:
            identifier = shell.make_identifier(ctxt.option.metavar)
        else:
            identifier = 'arg%d' % ctxt.option.get_positional_num()

        return '%s__%s' % (funcname, identifier)

    raise AssertionError('Should not be reached')


class GeneralHelpers:
    '''Class for including functions in the generation process.'''

    def __init__(self, config, prog, function_class):
        self.config = config
        self.function_prefix = config.function_prefix.replace('$PROG', prog)
        self.function_class = function_class

        # Dynamic functions
        self.dynamic_functions_code_to_funcname = {}

        # Builtin functins
        self.functions = {}
        self.used_functions = {}  # funcname:set(defines)
        self.global_defines = set()

        # Set debug mode
        if self.config.debug:
            self.define('DEBUG')

    # =========================================================================
    # Dynamic functions
    # =========================================================================

    def add_dynamic_func(self, ctxt, code):
        '''Add dynamically generated function.'''

        try:
            return self.dynamic_functions_code_to_funcname[code]
        except KeyError:
            funcname = self.get_dynamic_funcname(ctxt)
            self.dynamic_functions_code_to_funcname[code] = funcname
            return funcname

    def get_dynamic_funcname(self, ctxt):
        '''Return a unique function name for `ctxt`.'''

        funcname = make_completion_funcname_for_context(ctxt, self.function_prefix)
        num = 0
        funcname_plus_num = funcname
        while funcname_plus_num in self.dynamic_functions_code_to_funcname.values():
            funcname_plus_num = '%s%d' % (funcname, num)
            num += 1
        return funcname_plus_num

    def get_all_dynamic_functions(self):
        '''Return a list of code for all dynamically defined functions.'''

        r = []
        for code, funcname in self.dynamic_functions_code_to_funcname.items():
            r.append(self.function_class(funcname, code).get_code())
        return r

    # =========================================================================
    # Builtin functions
    # =========================================================================

    def add_function(self, function):
        '''Add a function to the function repository.'''

        assert isinstance(function, FunctionBase)

        self.functions[function.funcname] = function

    def define(self, name):
        '''Define a macro on global scope.'''

        self.global_defines.add(name)

    def use_function(self, function_name, *defines):
        '''Use a function from the function repository.'''

        if function_name not in self.functions:
            raise KeyError('No such function: %r' % function_name)

        if function_name not in self.used_functions:
            self.used_functions[function_name] = set(defines)
        else:
            self.used_functions[function_name].update(defines)

        # Add dependencies
        for function in self.functions[function_name].dependencies:
            self.use_function(function)

        return self.get_real_function_name(function_name)

    def get_real_function_name(self, function_name):
        '''Return the function name with its prefix.'''

        return '%s__%s' % (self.function_prefix, function_name)

    def is_used(self, function_name):
        '''Check if a function is used.'''

        return function_name in self.used_functions

    def get_used_functions_code(self):
        '''Return a list of code for all used functions.'''

        r = []

        for funcname, defines in self.used_functions.items():
            function = self.functions[funcname]
            realname = self.get_real_function_name(funcname)
            code = function.get_code(realname, self.global_defines | defines)

            for dep_func in function.dependencies:
                real_func = self.get_real_function_name(dep_func)
                code = replace_function_name(dep_func, real_func, code)

            code = code.replace('%PREFIX%', self.function_prefix)

            if not self.config.keep_comments:
                code = strip_comments(code)

            code = strip_double_empty_lines(code)

            r.append(code)

        r.extend(self.get_all_dynamic_functions())

        return r

    # =========================================================================
    # Other methods
    # =========================================================================

    def make_completion_funcname(self, commandline, suffix=''):
        '''Generates a function name for auto-completing a program or subcommand.

        Args:
            cmdline (CommandLine):
                The CommandLine instance representing the program or subcommand.

            suffix (str):
                The suffix that shall be appended to the result.

        Returns:
            str: The generated function name for auto-completion.

        This function is used to generate a unique function name for
        auto-completing a program or subcommand in the specified shell.
        It concatenates the names of all parent commandlines, including the
        current commandline, and converts them into a valid function name
        format.

        Example:
            For a program with the name 'my_program' and a subcommand with the name 'subcommand',
            the generated function name is '_my_program_subcommand'.
        '''
        assert isinstance(commandline, cli.CommandLine)

        commandlines = commandline.get_parents(include_self=True)
        prognames = [cmdline.prog for cmdline in commandlines]
        prognames[0] = self.function_prefix
        identifier = shell.make_identifier('_'.join(prognames))

        return f'{identifier}{suffix}'



#!/usr/bin/python3

import re
import os
import yaml

INFILE = 'messages.in.py'
OUTFILE = '../crazy_complete/messages.py'
TRANSLATIONS_INFILE = 'translations.yaml'

os.chdir(os.path.dirname(os.path.abspath(__file__)))


def get_fstring(translation, params):
    if not params:
        return f"'{translation}'"

    return f"f'{translation}'"


def make_function(args):
    msgid = args.pop('msgid')
    params = ', '.join(args.pop('parameters'))
    en = args.pop('en')

    r = f'def {msgid}({params}):\n'
    for lang, translation in args.items():
        r += f"    if LANG == '{lang}':\n"
        r += f"        return {get_fstring(translation, params)}\n"
    r += f"    return {get_fstring(en, params)}"
    return r


def make_functions():
    with open(TRANSLATIONS_INFILE, 'r', encoding='UTF-8') as fh:
        all_definitions = list(yaml.safe_load_all(fh))

    by_lang = {}

    for definition in all_definitions:
        by_lang[definition.pop('lang')] = definition

    specification = by_lang.pop('spec')

    for lang, defintions in by_lang.items():
        for key, translation in defintions.items():
            if not key in specification:
                raise Exception(f'{lang}: Invalid key: {key}')

    functions = []

    for msgid in sorted(specification.keys()):
        args = {'msgid': msgid, 'parameters': specification[msgid]}

        for lang, defintion in by_lang.items():
            try:
                args[lang] = defintion[msgid]
            except KeyError:
                raise Exception(f'{lang}: Missing translation: {msgid}')

        functions.append(make_function(args))

    return '\n\n\n'.join(functions)


with open(INFILE, 'r', encoding='UTF-8') as fh:
    content = fh.read()

with open(OUTFILE, 'w', encoding='UTF-8') as fh:
    fh.write(content)
    fh.write('\n')
    fh.write(make_functions())
    fh.write('\n')



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains utility functions.'''

import sys
from collections import namedtuple
from .type_utils import is_list_type


def warn(*a):
    '''Print a warning.'''

    print('WARNING:', *a, file=sys.stderr)


def print_err(*a):
    '''Print to STDERR.'''

    print(*a, file=sys.stderr)


class GeneralAbbreviationGenerator:
    '''A class for generating abbreviations from a list of words.

    Attributes:
        min_abbreviated_length (int):
            The minimum length of an abbreviation.

        abbreviations (dict):
            A dictionary mapping each word to its list of abbreviations.

        min_lengths (dict):
            A dictionary mapping each word to its minimum abbreviation length.
    '''

    def __init__(self, min_abbreviated_length, words):
        '''Initialize the GeneralAbbreviationGenerator instance.

        Args:
            min_abbreviated_length (int):
                The minimum length of an abbreviation.

            words (iterable of str):
                The list of words from which to generate abbreviations.
        '''

        assert isinstance(min_abbreviated_length, int)
        assert is_list_type(words)

        # pylint: disable=cell-var-from-loop

        self.min_abbreviated_length = min_abbreviated_length
        self.abbreviations = {}
        self.min_lengths = {}

        for word in words:
            self.abbreviations[word] = []
            self.min_lengths[word] = len(word)

            for length in range(len(word), self.min_abbreviated_length - 1, -1):
                abbrev = word[0:length]

                can_abbreviate = True
                for other_word in filter(lambda w: w != word, words):
                    if other_word.startswith(abbrev):
                        can_abbreviate = False
                        break

                if can_abbreviate or abbrev == word:
                    self.abbreviations[word].append(abbrev)
                    self.min_lengths[word] = length

    def get_abbreviations(self, word):
        '''Get the list of abbreviations for a given word.

        Args:
            word (str): The word for which to retrieve abbreviations.

        Returns:
            list: A list of abbreviations for the given word.
        '''

        assert isinstance(word, str)

        return self.abbreviations[word]

    def get_many_abbreviations(self, words):
        '''Get the list of abbreviations for multiple words.

        Args:
            words (iterable of str):
                The words for which to retrieve abbreviations.

        Returns:
            list: A list of abbreviations for the given words.
        '''

        assert is_list_type(words)

        r = []
        for word in words:
            r.extend(self.get_abbreviations(word))
        return r


class OptionAbbreviationGenerator(GeneralAbbreviationGenerator):
    '''AbbreviationGenerator for abbreviating long and old-style options.'''

    def __init__(self, words):
        assert is_list_type(words)

        words = list(words)

        for word in words:
            if word.startswith('--') and len(word) > 3:
                pass
            elif word.startswith('-') and len(word) > 2:
                pass
            else:
                raise ValueError('Not a long or old-style option: %r' % word)

        super().__init__(3, words)


class CommandAbbreviationGenerator(GeneralAbbreviationGenerator):
    '''AbbreviationGenerator for abbreviating commands.'''

    def __init__(self, words):
        super().__init__(1, words)


class DummyAbbreviationGenerator:
    '''A dummy abbreviation generator that returns the original word as the abbreviation.

    This class is used as a placeholder when abbreviation generation is not required.
    '''

    def __init__(self):
        pass

    def get_abbreviations(self, word):
        '''Don't abbreviate, just return the input.'''

        assert isinstance(word, str)

        return [word]

    def get_many_abbreviations(self, words):
        '''Don't abbreviate, just return the input.'''

        assert is_list_type(words)

        return words


def get_option_abbreviator(commandline):
    '''Return an OptionAbbreviationGenerator for options in `commandline`.'''

    if not commandline.abbreviate_options:
        return DummyAbbreviationGenerator()

    options = commandline.get_options(with_parent_options=commandline.inherit_options)
    option_strings = []

    for option in options:
        option_strings.extend(option.get_long_option_strings())
        option_strings.extend(option.get_old_option_strings())

    return OptionAbbreviationGenerator(option_strings)


def get_all_command_variations(commandline):
    '''Return all possible names for this command.

    If `commandline.abbreviate_commands` is True, also return abbreviated forms.
    '''

    if commandline.parent is None:
        return [commandline.prog] + commandline.aliases

    if commandline.abbreviate_commands:
        all_commands = []
        for subcommand in commandline.parent.subcommands.subcommands:
            all_commands.append(subcommand.prog)
        abbrevs = CommandAbbreviationGenerator(all_commands)
    else:
        abbrevs = DummyAbbreviationGenerator()

    return abbrevs.get_abbreviations(commandline.prog) + commandline.aliases


def get_defined_option_types(commandline):
    '''Return a tuple of defined option types.'''

    # pylint: disable=too-many-branches

    short_required = False
    short_optional = False
    short_flag     = False
    long_required  = False
    long_optional  = False
    long_flag      = False
    old_required   = False
    old_optional   = False
    old_flag       = False
    positionals    = False

    for cmdline in commandline.get_all_commandlines():
        if len(cmdline.get_positionals()) > 0:
            positionals = True

        if cmdline.get_subcommands() and len(cmdline.get_subcommands().subcommands) > 0:
            positionals = True

        for option in cmdline.options:
            if option.get_long_option_strings():
                if option.has_required_arg():
                    long_required = True
                elif option.has_optional_arg():
                    long_optional = True
                else:
                    long_flag = True

            if option.get_old_option_strings():
                if option.has_required_arg():
                    old_required = True
                elif option.has_optional_arg():
                    old_optional = True
                else:
                    old_flag = True

            if option.get_short_option_strings():
                if option.has_required_arg():
                    short_required = True
                elif option.has_optional_arg():
                    short_optional = True
                else:
                    short_flag = True

    return namedtuple('Types', (
        'positionals',
        'short_required', 'short_optional', 'short_flag', 'short',
        'long_required',  'long_optional',  'long_flag',  'long',
        'old_required',   'old_optional',   'old_flag',   'old'))(
        positionals,
        short_required,    short_optional,   short_flag,
        short_required or  short_optional or short_flag,
        long_required,     long_optional,    long_flag,
        long_required  or  long_optional  or long_flag,
        old_required,      old_optional,     old_flag,
        old_required   or  old_optional   or old_flag)


def get_query_option_strings(commandline, with_parent_options=True):
    '''Return a string that can be used by {fish,zsh}_query functions.

    Returns something like:
        "-f --flag -a= --argument= -o=? --optional=?"
    '''

    r = []

    for option in commandline.get_options(with_parent_options=with_parent_options):
        if option.has_required_arg():
            r.extend('%s=' % s for s in option.option_strings)
        elif option.has_optional_arg():
            r.extend('%s=?' % s for s in option.option_strings)
        else:
            r.extend(option.option_strings)

    return ' '.join(r)


def is_worth_a_function(commandline):
    '''Check if a commandline "is worth a function".

    This means that a commandline has one of:
        - Subcommands
        - Positionals (that have meaningful completion)
        - Options (that aren't --help or --version)
    '''

    if commandline.get_subcommands():
        return True

    for positional in commandline.get_positionals():
        if positional.complete[0] == 'none':
            pass
        elif positional.complete[0] in ('integer', 'float'):
            # integer/float without options have no real completion
            if len(positional.complete) > 1:
                return True
        else:
            return True

    options = commandline.get_options()
    options = list(filter(lambda o: '--help' not in o.option_strings and
                                    '--version' not in o.option_strings, options))
    if len(options) > 0:
        return True

    return False



#!/usr/bin/env python3

import os
import sys
import time
import json
import shlex
import yaml
import queue
import argparse
import threading
import subprocess
from collections import OrderedDict

from shells import BashShell, FishShell, ZshShell

# =============================================================================
# Script configuration
# =============================================================================

SHELLS              = ['bash', 'fish', 'zsh']
TMUX_SESSION_PREFIX = 'crazy-complete-test'
TESTS_INFILE        = 'tests.yaml'
TESTS_OUTFILE       = 'tests.new.yaml'
CRAZY_COMPLETE      = '../../crazy-complete'
COMPLETIONS_OUTDIR  = 'output'
TEMP_DIR            = 'tmp'
OUTPUT_WAIT         = 5

# =============================================================================
# Switch to the script's directory
# =============================================================================

os.chdir(os.path.dirname(os.path.abspath(__file__)))
CURRENT_DIRECTORY = os.path.abspath(os.getcwd())

# =============================================================================
# Commandline parser
# =============================================================================

argp = argparse.ArgumentParser()
argp.add_argument('-f', '--fast', action='store_true', default=False,
    help='Enable fast testing mode. For tests where the input matches the expected output, these tests will always pass')
argp.add_argument('-d', '--driver', default='tmux', choices=['pyte', 'tmux'],
    help='Select driver for tests')
argp.add_argument('-t', '--threads', default=1, type=int,
    help='Set the number of threads per shell')
argp.add_argument('-w', '--wait', default=OUTPUT_WAIT, type=float,
    help='Set how many seconds to wait for terminal output')
opts = argp.parse_args()

OUTPUT_WAIT = opts.wait

# =============================================================================
# Helper functions
# =============================================================================

def print_err(*args):
    print(*args, file=sys.stderr)

def run(args, env=None):
    result = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, env=env)

    if result.returncode != 0:
        raise Exception("Command %r failed: %s" % (args, result.stderr))

    return result.stdout

def indent(string, num_spaces):
    lines = string.split('\n')
    indented_lines = [((' ' * num_spaces) + line) if line.strip() else line for line in lines]
    return '\n'.join(indented_lines)

def make_yaml_block_string(s):
    return "|\n%s" % indent(s, 2)

# =============================================================================
# Import driver depending on command line arguments
# =============================================================================

if opts.driver == 'tmux':
    from tmux_driver import TmuxTerminal
elif opts.driver == 'pyte':
    print_err("The `pyte` driver has been disabled because it does not implement the `kitty`")
    print_err("protocol used by the Fish shell, causing tests to fail.")
    sys.exit(2)

# =============================================================================
# Ensure all dependencies are available
# =============================================================================

for shell in SHELLS:
    try:
        run(['sh', '-c', f'type {shell}'])
    except Exception:
        print_err(f'Program `{shell}` not found')
        sys.exit(2)

if not os.path.exists('/usr/share/bash-completion/bash_completion'):
    print_err('File `/usr/share/bash-completion/bash_completion` not found. Is `bash-completion` installed?')
    sys.exit(2)

# =============================================================================
# Test code
# =============================================================================

class Test:
    def __init__(self, data):
        self.number = data.get('number', None) 
        self.definition_file = data['definition_file']
        self.description = data['description']
        self.comment = data.get('comment', None)
        self.send = data['send']
        self.require_order = data.get('require_order', False)
        self.bash_tabs = data.get('bash_tabs', 1)
        self.fish_tabs = data.get('fish_tabs', 1)
        self.zsh_tabs = data.get('zsh_tabs', 1)
        self.bash_expected = data.get('bash_expected', '').strip()
        self.fish_expected = data.get('fish_expected', '').strip()
        self.zsh_expected = data.get('zsh_expected', '').strip()
        self.bash_result = None
        self.fish_result = None
        self.zsh_result = None

    def get_tabs(self, shell):
        if shell == 'bash':
            return self.bash_tabs
        if shell == 'fish':
            return self.fish_tabs
        if shell == 'zsh':
            return self.zsh_tabs
        raise

    def get_expected(self, shell):
        if shell == 'bash':
            return self.bash_expected
        if shell == 'fish':
            return self.fish_expected
        if shell == 'zsh':
            return self.zsh_expected

    def get_result(self, shell):
        if shell == 'bash':
            return self.bash_result
        if shell == 'fish':
            return self.fish_result
        if shell == 'zsh':
            return self.zsh_result
        raise

    def set_result(self, shell, result):
        if shell == 'bash':
            self.bash_result = result
        elif shell == 'fish':
            self.fish_result = result
        elif shell == 'zsh':
            self.zsh_result = result
        else:
            raise

    def passed(self, shell):
        result = self.get_result(shell)
        expected = self.get_expected(shell)

        if self.require_order:
            return result.split() == expected.split()
        else:
            return sorted(result.split()) == sorted(expected.split())

    def to_yaml(self):
        r = OrderedDict()

        r['number']          = str(self.number)
        r['definition_file'] = json.dumps(self.definition_file)
        r['description']     = json.dumps(self.description)
        if self.comment:
            r['comment']     = json.dumps(self.comment)
        r['send']            = json.dumps(self.send)
        if self.require_order:
            r['require_order'] = 'true'
        if self.bash_tabs != 1:
            r['bash_tabs']   = str(self.bash_tabs)
        r['bash_expected']   = make_yaml_block_string(self.bash_result)
        if self.fish_tabs != 1:
            r['fish_tabs']   = str(self.fish_tabs)
        r['fish_expected']   = make_yaml_block_string(self.fish_result)
        if self.zsh_tabs != 1:
            r['zsh_tabs']    = str(self.zsh_tabs)
        r['zsh_expected']    = make_yaml_block_string(self.zsh_result)

        return '\n'.join('%s: %s' % key_value for key_value in r.items())


class Tests:
    def __init__(self, tests):
        self.definition_files = tests.pop(0)
        self.tests = [Test(data) for data in tests]

    def enumerate_tests(self):
        number = 1
        for test in self.tests:
            test.number = number
            number += 1

    def find_test_by_number(self, num):
        for test in self.tests:
            if test.number == num:
                return test

        raise Exception(f"No test with number {num} found")

    def write_tests_file(self, outfile):
        r = []

        files = ''
        for file, args in self.definition_files.items():
            files += '%s: %s\n' % (file, json.dumps(args))
        r += [files.strip()]

        for test in self.tests:
            r += [test.to_yaml()]

        r = '\n---\n'.join(r)

        with open(outfile, 'w') as fh:
            fh.write(r)

    def generate_completion_files(self):
        try:
            os.mkdir(COMPLETIONS_OUTDIR)
        except FileExistsError:
            if not os.path.isdir(COMPLETIONS_OUTDIR):
                raise NotADirectoryError(COMPLETIONS_OUTDIR) from None

        try:
            os.mkdir(TEMP_DIR)
        except FileExistsError:
            if not os.path.isdir(TEMP_DIR):
                raise NotADirectoryError(TEMP_DIR) from None

        batch_file = '%s/%s' % (TEMP_DIR, 'batch.txt')
        batch_commands = []

        for file, args in self.definition_files.items():
            definition_file = args['definition_file']
            temp_definition_file = f'{TEMP_DIR}/{definition_file}'

            with open(definition_file, 'r', encoding='UTF-8') as fh:
                content = fh.read()
                content = content.replace('%TEST_DIR%', CURRENT_DIRECTORY)

            with open(temp_definition_file, 'w', encoding='UTF-8') as fh:
                fh.write(content)

            for shell in SHELLS:
                cmd =  ['--debug', '--zsh-compdef=False']
                for arg in args['args']:
                    cmd += [arg.replace('$shell', shell)]
                cmd += ['-o', f'{COMPLETIONS_OUTDIR}/{file}.{shell}']
                cmd += [shell, temp_definition_file]
                cmd = ' '.join(shlex.quote(arg) for arg in cmd)
                batch_commands.append(cmd)

        with open(batch_file, 'w', encoding='utf-8') as fh:
            print_err("Writing batch file '%s' ..." % batch_file)
            fh.write('\n'.join(batch_commands))

        print_err('Running batch file ...')
        args = [CRAZY_COMPLETE, '--batch', batch_file]
        subprocess.run(args, check=True)
        print_err('Finished running batch file ...')

def tests_worker_thread(thread_id, shell, input_queue, result_queue):
    if opts.driver == 'tmux':
        term = TmuxTerminal(f'{TMUX_SESSION_PREFIX}-{shell}-{thread_id}')
    elif opts.driver == 'pyte':
        term = PyteTerminal()

    term_shell = {
        'bash': BashShell,
        'fish': FishShell,
        'zsh':  ZshShell
    }[shell](term)

    old_definition_file = None

    while True:
        try:
            test = input_queue.get_nowait()
        except queue.Empty:
            break

        if old_definition_file != test.definition_file:
            old_definition_file = test.definition_file

            try:
                term_shell.stop()
            except Exception:
                pass
            completion_file = '%s/%s.%s' % (COMPLETIONS_OUTDIR, test.definition_file, shell)
            term_shell.start()
            term.resize_window(80, 100)
            term_shell.set_prompt()
            term_shell.init_completion()
            term_shell.load_completion(completion_file)
            time.sleep(1.5)

        output = term.complete(
            test.send,
            test.get_tabs(shell),
            OUTPUT_WAIT,
            test.get_expected(shell),
            opts.fast)

        result = {
            'number': test.number,
            'shell':  shell,
            'result': output
        }

        result_queue.put(result)

    try:
        term_shell.stop()
    except Exception:
        pass

class Tester():
    def __init__(self, tests):
        self.tests = tests
        self.result_queue = queue.Queue()
        self.input_queues = {}
        self.threads = []
        self.num_failed = 0

    def run(self):
        for shell in SHELLS:
            self.input_queues[shell] = queue.Queue()
            for test in self.tests.tests:
                self.input_queues[shell].put(test)

        for thread_id in range(opts.threads):
            for shell in SHELLS:
                thread = threading.Thread(
                    target=tests_worker_thread,
                    args=(thread_id, shell, self.input_queues[shell], self.result_queue)
                )

                thread.start()
                self.threads.append(thread)

        while self.threads_are_running():
            self.eat_queue()
            time.sleep(0.1)
        self.eat_queue()

    def threads_are_running(self):
        for thread in self.threads:
            if thread.is_alive():
                return True
        return False

    def eat_queue(self):
        while not self.result_queue.empty():
            self.process_result(self.result_queue.get())

    def process_result(self, result):
        test = self.tests.find_test_by_number(result['number'])
        shell = result['shell']
        test.set_result(shell, result['result'])

        if not test.passed(shell):
            self.num_failed += 1
            print_err("Test #%02d (%-4s - %s) failed" % (test.number, shell, test.description))
        else:
            print_err("Test #%02d (%-4s - %s) OK" % (test.number, shell, test.description))

# =============================================================================
# Main
# =============================================================================

with open(TESTS_INFILE, 'r') as fh:
    tests = Tests(list(yaml.safe_load_all(fh)))

tests.enumerate_tests()
tests.generate_completion_files()
tester = Tester(tests)
tester.run()
tests.write_tests_file(TESTS_OUTFILE)
if tester.num_failed:
    print_err(f'{tester.num_failed} tests failed.')
    print_err(f'Use diff or vimdiff on `{TESTS_INFILE}` and `{TESTS_OUTFILE}` for further details')
    if opts.threads > 1:
        print_err('NOTE:')
        print_err(' A high value for -t|--threads may cause that tests fail.')
        print_err(' Consider running this script again with `-t 1`.')
    sys.exit(1)
else:
    print_err("All tests passed")
    sys.exit(0)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains functions for generating a YAML file from help texts.'''

from . import help_parser
from . import yaml_source
from .str_utils import indent


def fix_description(s):
    '''Fix description.'''

    s = s.strip()

    # Replace hyphens followed by a newline ('-\n') with a simple hyphen.
    # This prevents words like "Some-word-with-\nhyphens" from being split
    # incorrectly due to the newline. Instead, it will correctly join as
    # "Some-word-with-hyphens".
    s = s.replace('-\n', '-')

    s = s.replace('\n', ' ')
    return s


def strip_metavar(s):
    '''Strip metavar.'''

    if s and s[0] == '[' and s[-1] == ']':
        s = s[1:-1]

    if s and s[0] == '=':
        s = s[1:]

    if s and s[0] == '<' and s[-1] == '>':
        s = s[1:-1]

    return s


def complete_for_metavar(string):
    '''Return completion for metavar.'''

    string = string.strip('<>')
    lower = string.lower()

    equals = {
        'f':            ['file'],
        'n':            ['integer'],
        'duration':     ['integer'],
        'bits':         ['integer'],
        'cols':         ['integer'],
        'column':       ['integer'],
        'digits':       ['integer'],
        'size':         ['integer'],
        'bytes':        ['integer'],
        'pid':          ['pid'],
        'user':         ['user'],
        'group':        ['group'],
        'program':      ['command'],
        'command':      ['command'],
        'cmd':          ['cmd'],
        'signal':       ['signal'],
        'sig':          ['signal'],
    }

    endswith = {
        'file':         ['file'],
        'filename':     ['file'],
        'dir':          ['directory'],
        'directory':    ['directory'],
        'seconds':      ['integer'],
        'length':       ['integer'],
        'width':        ['integer'],
        'num':          ['integer'],
        'number':       ['integer'],
    }

    if lower in equals:
        return equals[lower]

    for suffix, complete in endswith.items():
        if lower.endswith(suffix):
            return complete

    if string.startswith('{') and string.endswith('}'):
        string = string.strip('{}')
        for sep in ['|', ',']:
            if sep in string:
                return ['choices', [item.strip() for item in string.split(sep)]]

    return ['none']


def from_file_to_yaml(file):
    '''Parse help text in a file and return YAML.'''

    with open(file, 'r', encoding='utf-8') as fh:
        content = fh.read()

    prog = help_parser.get_program_name_from_help(content)
    char_stream = help_parser.CharStream(content)
    parsed = help_parser.parse(char_stream)

    output = []

    output.append(f'prog: "{prog}"\nhelp: "<Program description here>"\noptions:')

    for obj in parsed:
        if isinstance(obj, help_parser.Unparsed):
            output.append(f"# {obj.text.rstrip()}")
        elif isinstance(obj, help_parser.OptionsWithDescription):
            option_dict = {
                'option_strings': [],
                'metavar':        None,
                'optional_arg':   False,
                'help':           fix_description(obj.description or '')
            }

            for option in obj.options:
                option_dict['option_strings'].append(option.option)

                if option.optional:
                    option_dict['optional_arg'] = True

                if option.metavar:
                    option_dict['metavar'] = strip_metavar(option.metavar)
                    option_dict['complete'] = complete_for_metavar(option_dict['metavar'])

            output.append(
                indent(yaml_source.option_to_yaml(option_dict), 2)
            )

            output.append('')

    return '\n'.join(output)



import time

SHELL_ENV = {
    # disable readline's rcfile for bash
    'INPUTRC': '/dev/null',

    # set locale to C to ensure same ordering of options
    'LANG':              'C',
    'LC_CTYPE':          'C',
    'LC_NUMERIC':        'C',
    'LC_TIME':           'C',
    'LC_COLLATE':        'C',
    'LC_MONETARY':       'C',
    'LC_MESSAGES':       'C',
    'LC_PAPER':          'C',
    'LC_NAME':           'C',
    'LC_ADDRESS':        'C',
    'LC_TELEPHONE':      'C',
    'LC_MEASUREMENT':    'C',
    'LC_IDENTIFICATION': 'C',
    'LC_ALL':            'C'
}

def wait_for_prompt(terminal):
    terminal.wait_for_text('>', 5, 0.1)

class ShellBase:
    def __init__(self, term):
        self.term = term

    def stop(self):
        self.term.stop()

class BashShell(ShellBase):
    def start(self):
        self.term.start(['bash', '--norc', '--noprofile', '+o', 'history'], SHELL_ENV)

    def set_prompt(self):
        self.term.send_line("clear; PS1='> '")
        wait_for_prompt(self.term)

    def init_completion(self):
        self.term.send_line('source /usr/share/bash-completion/bash_completion')
        self.term.send_line('bind "set show-all-if-ambiguous on"')
        self.term.send_line('bind "set show-all-if-unmodified on"')
        self.term.send_line('clear')
        wait_for_prompt(self.term)

    def load_completion(self, file):
        self.term.send_line('clear; source %s' % file)
        wait_for_prompt(self.term)

class FishShell(ShellBase):
    def start(self):
        self.term.start(['fish', '--no-config'], SHELL_ENV | {'fish_autosuggestion_enabled': 0})

    def set_prompt(self):
        self.term.send_line("clear; function fish_prompt; printf '> '; end")
        wait_for_prompt(self.term)

    def init_completion(self):
        pass

    def load_completion(self, file):
        self.term.send_line('clear; source %s' % file)
        wait_for_prompt(self.term)

class ZshShell(ShellBase):
    def start(self):
        self.term.start(['zsh', '--no-rcs'], SHELL_ENV)

    def set_prompt(self):
        self.term.send_line("clear; PROMPT='> '")
        wait_for_prompt(self.term)

    def init_completion(self):
        self.term.send_line('clear; autoload -U compinit && compinit')
        wait_for_prompt(self.term)

    def load_completion(self, file):
        self.term.send_line('clear; source %s' % file)
        wait_for_prompt(self.term)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Code for parsing a command line in Bash, version 2'''

from collections import namedtuple

from . import utils
from . import shell
from . import bash_patterns
from .str_utils import indent
from .bash_parser_subcommand_code import (
    make_subcommand_switch_code, get_subcommand_path
)
from .preprocessor import preprocess

_PARSER_CODE = '''\
#ifdef positionals
POSITIONALS=()
#endif
END_OF_OPTIONS=0

local cmd="root" argi arg i char trailing_chars VAR ARGS

%FIND_OPTION_CODE%

__append_to_array() {
  local -n arr=$1
  arr+=("$2")
}

for ((argi=1; argi < cword; ++argi)); do
  arg="${words_dequoted[argi]}"

  case "$arg" in
    --)
      END_OF_OPTIONS=1
#ifdef positionals
      POSITIONALS+=("${words_dequoted[@]:$((++argi))}")
#endif
      return;;
#ifdef long_options
    --*=*)
      if __find_option "$cmd" "${arg%%=*}"
      then __append_to_array "$VAR" "${arg#*=}"
      fi;;
    --*)
      if __find_option "$cmd" "$arg"; then
        if [[ "$ARGS" == 1 ]]
        then __append_to_array "$VAR" "${words_dequoted[++argi]}"
        else __append_to_array "$VAR" "_OPT_ISSET_"
        fi
      fi;;
#else
    --*);;
#endif
    -?*) # ignore '-'
#ifdef old_options
      if [[ "$arg" == -*=* ]]; then
        if __find_option "$cmd" "${arg%%=*}"; then
          __append_to_array "$VAR" "${arg#*=}"
          continue
        fi
      fi

      if __find_option "$cmd" "$arg"; then
        if [[ "$ARGS" == 1 ]]
        then __append_to_array "$VAR" "${words_dequoted[++argi]}"
        else __append_to_array "$VAR" "_OPT_ISSET_"
        fi

        continue
      fi
#endif

#ifdef short_options
      for ((i=1; i < ${#arg}; ++i)); do
        char="${arg:$i:1}"
        trailing_chars="${arg:$((i + 1))}"

        if __find_option "$cmd" "-$char"; then
          if [[ "$ARGS" == 1 ]]; then
            if [[ -n "$trailing_chars" ]]
            then __append_to_array "$VAR" "$trailing_chars"
            else __append_to_array "$VAR" "${words_dequoted[++argi]}"
            fi
            break;
#ifdef short_optionals
          elif [[ "$ARGS" == '?' ]]; then
            if [[ -n "$trailing_chars" ]]
            then __append_to_array "$VAR" "$trailing_chars"
            else __append_to_array "$VAR" _OPT_ISSET_
            fi
            break;
#endif
          else
            __append_to_array "$VAR" "_OPT_ISSET_"
          fi
        fi
      done
#endif
      ;;
#ifdef positionals
    *)
      POSITIONALS+=("$arg")
%SUBCOMMAND_SWITCH_CODE%
      ;;
#endif
  esac
done
#ifdef positionals

for ((; argi <= cword; ++argi)); do
  case "${words_dequoted[argi]}" in
    -?*);;
    *) POSITIONALS+=("${words_dequoted[argi]}");;
  esac
done
#endif'''

_OPT_ISSET = '_OPT_ISSET_'


def _make_find_option_code(commandline, variable_manager):
    cmdlines = list(reversed(commandline.get_all_commandlines()))
    code = []

    if len(cmdlines) == 1:
        code += [_make_cmd_plus_option_switch_code(cmdlines[0], variable_manager, True)]
    else:
        for cmdline in cmdlines:
            code += [_make_cmd_plus_option_switch_code(cmdline, variable_manager)]

    r = '__find_option() {\n'
    r += '%s\n' % indent('\n'.join(filter(None, code)), 2)
    r += '  return 1\n'
    r += '}'
    return r


def _make_cmd_plus_option_switch_code(commandline, variable_manager, omit_cmd_check=False):
    option_cases = _generate_option_cases(commandline, variable_manager)
    if not option_cases:
        return None

    if omit_cmd_check:
        return _make_option_switch_code(option_cases)

    command = get_subcommand_path(commandline)
    command = shell.quote(command)
    if commandline.inherit_options:
        command += '*'

    r = f'case "$1" in {command})\n'
    r += '%s\n' % indent(_make_option_switch_code(option_cases), 2)
    r += 'esac'
    return r


def _make_option_switch_code(option_cases):
    r = 'case "$2" in\n'

    for case in option_cases:
        pattern = bash_patterns.make_pattern(case.option_strings)
        r += f'  {pattern}) VAR={case.variable}; ARGS={case.args}; return;;\n'

    r += 'esac'
    return r


def generate(commandline, variable_manager):
    '''Generate code for parsing the command line.'''

    find_option_code = _make_find_option_code(commandline, variable_manager)
    subcommand_switch_code = make_subcommand_switch_code(commandline)

    defines = []
    types = utils.get_defined_option_types(commandline)
    if types.short:
        defines.append('short_options')
    if types.long:
        defines.append('long_options')
    if types.old:
        defines.append('old_options')
    if types.short_optional:
        defines.append('short_optionals')
    if types.positionals:
        defines.append('positionals')

    s = preprocess(_PARSER_CODE, defines)

    if subcommand_switch_code:
        s = s.replace('%SUBCOMMAND_SWITCH_CODE%',
                      indent(subcommand_switch_code, 6))
    else:
        s = s.replace('%SUBCOMMAND_SWITCH_CODE%\n', '')

    if find_option_code:
        s = s.replace('%FIND_OPTION_CODE%', find_option_code)
    else:
        s = s.replace('%FIND_OPTION_CODE%\n', '')

    return s


def _generate_option_cases(commandline, variable_manager):
    OptionCase = namedtuple('OptionCase', ['option_strings', 'variable', 'args'])
    options = commandline.get_options()
    abbreviations = utils.get_option_abbreviator(commandline)
    option_cases = []

    for option in options:
        option_strings = option.get_short_option_strings()

        option_strings += abbreviations.get_many_abbreviations(
                option.get_long_option_strings())

        option_strings += abbreviations.get_many_abbreviations(
                option.get_old_option_strings())

        value_variable = variable_manager.capture_variable(option)

        if option.has_optional_arg():
            args = "'?'"
        elif option.has_required_arg():
            args = '1'
        else:
            args = '0'

        option_cases.append(OptionCase(option_strings, value_variable, args))

    return option_cases



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''This module contains the CommandLine, Option and Positional classes.'''

from collections import OrderedDict
from types import NoneType

from .type_utils import validate_type
from .errors import CrazyError, CrazyTypeError
from .str_utils import (
    contains_space, is_valid_option_string,
    is_valid_variable_name, is_empty_or_whitespace
)

# pylint: disable=redefined-builtin
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments


class ExtendedBool:
    '''Class that holds an extended bool.'''

    TRUE    = True
    FALSE   = False
    INHERIT = 'INHERIT'


def is_extended_bool(obj):
    '''Check if `obj` an instance of `ExtendedBool`.'''

    return obj in (True, False, ExtendedBool.INHERIT)


class CommandLine:
    '''Represents a command line with options, positionals, and subcommands.'''

    def __init__(self,
                 prog,
                 parent=None,
                 help=None,
                 aliases=None,
                 wraps=None,
                 abbreviate_commands=ExtendedBool.INHERIT,
                 abbreviate_options=ExtendedBool.INHERIT,
                 inherit_options=ExtendedBool.INHERIT):
        '''Initializes a CommandLine object with the specified parameters.

        Args:
            prog (str):
                The name of the program (or subcommand).

            help (str):
                The help message for the program (or subcommand).

            parent (CommandLine or None):
                The parent command line object, if any.

            aliases (list of str or None):
                Aliases for this command.

            wraps (str or None):
                Specify the command to inherit options from.

            abbreviate_commands (ExtendedBool):
                Specifies if commands can be abbreviated.

            abbreviate_options (ExtendedBool):
                Specifies if options can be abbreviated.

            inherit_options (ExtendedBool):
                Specifies if options are visible to subcommands.
        '''
        if aliases is None:
            aliases = []

        validate_type(prog, (str,), 'prog')
        validate_type(parent, (CommandLine, NoneType), 'parent')
        validate_type(help, (str, NoneType), 'help')
        validate_type(aliases, (list,), 'aliases')
        validate_type(wraps, (str, NoneType), 'wraps')

        if wraps is not None:
            if is_empty_or_whitespace(wraps):
                raise CrazyError('wraps is empty')

            if contains_space(wraps):
                raise CrazyError('wraps: cannot contain space')

        for index, alias in enumerate(aliases):
            if not isinstance(alias, str):
                raise CrazyTypeError(f'aliases[{index}]', 'str', alias)

            if contains_space(alias):
                raise CrazyError(f'aliases[{index}]: cannot contain space')

        if not is_extended_bool(abbreviate_commands):
            raise CrazyTypeError('abbreviate_commands', 'ExtendedBool', abbreviate_commands)

        if not is_extended_bool(abbreviate_options):
            raise CrazyTypeError('abbreviate_options', 'ExtendedBool', abbreviate_options)

        if not is_extended_bool(inherit_options):
            raise CrazyTypeError('inherit_options', 'ExtendedBool', inherit_options)

        self.prog = prog
        self.parent = parent
        self.help = help
        self.aliases = aliases
        self.wraps = wraps
        self.abbreviate_commands = abbreviate_commands
        self.abbreviate_options = abbreviate_options
        self.inherit_options = inherit_options
        self.options = []
        self.positionals = []
        self.subcommands = None

    def add_option(self, option_strings, **parameters):
        '''Adds a new option to the command line.

        For a list of valid parameters, see `class Option`.

        Returns:
            Option: The newly added Option object.
        '''

        o = Option(self, option_strings, **parameters)
        self.options.append(o)
        return o

    def add_positional(self, number, **parameters):
        '''Adds a new positional argument to the command line.

        For a list of valid parameters, see `class Positional`.

        Returns:
            Positional: The newly added Positional object.
        '''

        p = Positional(self, number, **parameters)
        self.positionals.append(p)
        return p

    def add_mutually_exclusive_group(self, group):
        '''Adds a new mutually exclusive group.

        Returns:
            MutuallyExclusiveGroup: The newly created mutually exclusive group.
        '''

        return MutuallyExclusiveGroup(self, group)

    def add_subcommands(self):
        '''Adds a subcommands option to the command line.

        Returns:
            SubCommandsOption: The newly created subcommands option.

        Raises:
            CrazyError: If the command line object already has subcommands.
        '''

        if self.subcommands:
            raise CrazyError('CommandLine object already has subcommands')

        self.subcommands = SubCommandsOption(self)
        return self.subcommands

    class OptionsGetter:
        '''A class for getting options from a CommandLine object.'''

        def __init__(self, commandline, with_parent_options=False, only_with_arguments=False):
            self.options = OrderedDict()

            if with_parent_options:
                commandlines = commandline.get_parents(include_self=True)
            else:
                commandlines = [commandline]

            for cmdline in reversed(commandlines):
                for option in cmdline.options:
                    if not only_with_arguments or option.complete:
                        self._add(option, cmdline)

        def _add(self, option, commandline):
            key = option.get_option_strings_key()

            if key not in self.options:
                self.options[key] = (commandline, [])

            if self.options[key][0] == commandline:
                self.options[key][1].append(option)

        def get(self):
            '''Return the result.'''
            r = []
            for _, options in self.options.values():
                for option in options:
                    r.append(option)
            return r

    def get_options(self, with_parent_options=False, only_with_arguments=False):
        '''Gets a list of options associated with the command line.

        Args:
            with_parent_options (bool):
                If True, include options from parent command lines.

            only_with_arguments (bool):
                If True, include only options that take arguments.

        Returns:
            list: A list of Option objects
        '''

        assert isinstance(with_parent_options, bool)
        assert isinstance(only_with_arguments, bool)

        getter = CommandLine.OptionsGetter(
            self,
            with_parent_options=with_parent_options,
            only_with_arguments=only_with_arguments)

        return getter.get()

    def get_option_strings(self, with_parent_options=False, only_with_arguments=False):
        '''Gets a list of option strings associated with the command line.

        Args:
            with_parent_options (bool):
                If True, include options from parent command lines.

            only_with_arguments (bool):
                If True, include only options that take arguments.

        Returns:
            list: A list of option strings
        '''

        assert isinstance(with_parent_options, bool)
        assert isinstance(only_with_arguments, bool)

        option_strings = []

        options = self.get_options(
            with_parent_options=with_parent_options,
            only_with_arguments=only_with_arguments)

        for o in options:
            option_strings.extend(o.option_strings)

        return option_strings

    def get_final_options(self):
        '''Gets a list of all final options.'''

        r = []
        cmdline = self

        while cmdline:
            for option in cmdline.options:
                if option.final:
                    r.append(option)
            cmdline = cmdline.parent

        return r

    def get_final_option_strings(self):
        '''Gets a list of all final option strings.'''

        r = []
        for option in self.get_final_options():
            r.extend(option.option_strings)
        return r

    def get_positionals(self):
        '''Gets a list of positional arguments associated with the command line.

        Note:
            SubCommandsOption objects are not considered positional arguments
            and are not included in the list.

        Returns:
            list: A list of positional arguments
        '''

        return list(self.positionals)

    def get_subcommands(self):
        '''Gets the subcommands of the command line.

        Returns:
            SubCommandsOption or None:
                The subcommands option if it exists, otherwise None.
        '''

        return self.subcommands

    def get_parents(self, include_self=False):
        '''Gets a list of parent CommandLine objects.

        Args:
            include_self (bool):
                If True, includes the current CommandLine object in the list.

        Returns:
            list: A list of parent CommandLine objects.
        '''

        assert isinstance(include_self, bool)

        parents = []

        parent = self.parent
        while parent:
            parents.insert(0, parent)
            parent = parent.parent

        if include_self:
            parents.append(self)

        return parents

    def get_root_commandline(self):
        '''Return the root commandline.'''

        if not self.parent:
            return self

        return self.parent.get_root_commandline()

    def get_options_by_option_strings(self, option_strings):
        '''Return all options containing one option_strings.'''

        result = []

        for option_string in option_strings:
            found = False
            for option in self.options:
                if option_string in option.option_strings:
                    if option not in result:
                        result.append(option)
                    found = True
                    break
            if not found:
                raise CrazyError('Option %r not found' % option_string)

        return result

    def get_highest_positional_num(self):
        '''Get the highest positional number.'''

        highest = 0
        for positional in self.positionals:
            highest = max(highest, positional.number)
        if self.subcommands:
            highest += 1
        return highest

    def get_program_name(self):
        '''Return the program name.'''

        commandlines = self.get_parents(include_self=True)
        return commandlines[0].prog

    def get_command_path(self, progname=None):
        '''Return the full command path.'''
        prognames = [c.prog for c in self.get_parents(include_self=True)]

        if progname is not None:
            prognames[0] = progname

        return ' '.join(prognames)

    def visit_commandlines(self, callback):
        '''Apply a callback to all CommandLine objects.'''

        callback(self)
        if self.get_subcommands():
            for sub in self.get_subcommands().subcommands:
                sub.visit_commandlines(callback)

    def get_all_commandlines(self):
        '''Get a list of all defined commandlines.'''

        result = []
        self.visit_commandlines(result.append)
        return result

    def copy(self):
        '''Make a deep copy of the current CommandLine object.'''

        copy = CommandLine(
            self.prog,
            parent              = None,
            help                = self.help,
            aliases             = self.aliases,
            wraps               = self.wraps,
            abbreviate_commands = self.abbreviate_commands,
            abbreviate_options  = self.abbreviate_options,
            inherit_options     = self.inherit_options)

        for option in self.options:
            copy.add_option(
                option.option_strings,
                metavar         = option.metavar,
                help            = option.help,
                complete        = option.complete,
                nosort          = option.nosort,
                optional_arg    = option.optional_arg,
                groups          = option.groups,
                repeatable      = option.repeatable,
                final           = option.final,
                hidden          = option.hidden,
                when            = option.when,
                capture         = option.capture,
                long_opt_arg_sep = option.long_opt_arg_sep
            )

        for positional in self.positionals:
            copy.add_positional(
                positional.number,
                metavar         = positional.metavar,
                help            = positional.help,
                repeatable      = positional.repeatable,
                nosort          = positional.nosort,
                complete        = positional.complete,
                when            = positional.when,
                capture         = positional.capture
            )

        if self.subcommands is not None:
            subcommands_option = copy.add_subcommands()
            for subparser in self.subcommands.subcommands:
                subcommands_option.add_commandline_object(subparser.copy())

        return copy

    def __eq__(self, other):
        return (
            isinstance(other, CommandLine)                        and
            self.prog                == other.prog                and
            self.aliases             == other.aliases             and
            self.help                == other.help                and
            self.wraps               == other.wraps               and
            self.abbreviate_commands == other.abbreviate_commands and
            self.abbreviate_options  == other.abbreviate_options  and
            self.inherit_options     == other.inherit_options     and
            self.options             == other.options             and
            self.positionals         == other.positionals         and
            self.subcommands         == other.subcommands
        )

    def __repr__(self):
        r = '{'
        r += f'\nprog: {self.prog!r},'
        r += f'\naliases: {self.aliases!r},'
        r += f'\nwraps: {self.wraps!r},'
        r += f'\nhelp: {self.help!r},'
        r += f'\nabbreviate_commands: {self.abbreviate_commands!r},'
        r += f'\nabbreviate_options: {self.abbreviate_options!r},'
        r += f'\ninherit_options: {self.inherit_options!r},'
        r += f'\noptions: {self.options!r},'
        r += f'\npositionals: {self.positionals!r},'
        r += f'\nsubcommands: {self.subcommands!r},'
        r += '\n}'
        return r


class Positional:
    '''Class representing a command line positional.'''

    def __init__(
            self,
            parent,
            number,
            metavar=None,
            help=None,
            complete=None,
            nosort=False,
            repeatable=False,
            when=None,
            capture=None):
        '''Initializes a Positional object with the specified parameters.

        Args:
            parent (CommandLine):
                The parent command line object, if any.

            number (int):
                The number of the positional argument (starting from 1)

            metavar (str):
                The metavar for the positional.

            help (str):
                The help message for the positional.

            repeatable (bool):
                Specifies if positional can be specified more times

            complete (list):
                The completion specification for the positional.

            nosort (bool):
                Do not sort the completion suggestions.

            when (str):
                Specifies a condition for showing this positional.

            capture (str):
                Specifies the variable name for capturing
        '''

        validate_type(parent, (CommandLine, NoneType), 'parent')
        validate_type(number, (int,), 'number')
        validate_type(metavar, (str, NoneType), 'metavar')
        validate_type(help, (str, NoneType), 'help')
        validate_type(complete, (list, tuple, NoneType), 'complete')
        validate_type(repeatable, (bool,), 'repeatable')
        validate_type(nosort, (bool,), 'nosort')
        validate_type(when, (str, NoneType), 'when')
        validate_type(capture, (str, NoneType), 'capture')

        if capture is not None and not is_valid_variable_name(capture):
            raise CrazyError(f"Invalid variable name: {capture!r}")

        if number <= 0:
            raise CrazyError(f'number: value ({number}) is invalid, number has to be >= 1')

        self.parent = parent
        self.number = number
        self.metavar = metavar
        self.help = help
        self.repeatable = repeatable
        self.complete = complete if complete else ['none']
        self.nosort = nosort
        self.when = when
        self.capture = capture

    def get_positional_index(self):
        '''Returns the index of the current positional argument within the
        current commandline, including parent commandlines.

        Returns:
            int: The index of the positional argument.
        '''

        positional_no = self.number - 1

        for commandline in self.parent.get_parents():
            highest = 0
            for positional in commandline.get_positionals():
                highest = max(highest, positional.number)
            positional_no += highest

            if commandline.get_subcommands():
                positional_no += 1

        return positional_no

    def get_positional_num(self):
        '''Returns the number of the current positional argument within the
        current commandline, including parent commandlines.

        Note:
            This is the same as `CommandLine.get_positional_index() + 1`.

        Returns:
            int: The number of the positional argument.
        '''

        return self.get_positional_index() + 1

    def __eq__(self, other):
        return (
            isinstance(other, Positional)       and
            self.number     == other.number     and
            self.metavar    == other.metavar    and
            self.help       == other.help       and
            self.repeatable == other.repeatable and
            self.complete   == other.complete   and
            self.nosort     == other.nosort     and
            self.when       == other.when       and
            self.capture    == other.capture
        )


def is_short_option_string(string):
    '''Return True if string is a short option string.'''
    return len(string) == 2


def is_long_option_string(string):
    '''Return True if string is a long option string.'''
    return string.startswith('--')


def is_old_option_string(string):
    '''Return True if string is an old option string.'''
    return not string.startswith('--') and len(string) > 2


class Option:
    '''Class representing a command line option.'''

    # pylint: disable=too-many-locals

    def __init__(
            self,
            parent,
            option_strings,
            metavar=None,
            help=None,
            complete=None,
            nosort=False,
            groups=None,
            optional_arg=False,
            repeatable=ExtendedBool.INHERIT,
            final=False,
            hidden=False,
            when=None,
            capture=None,
            long_opt_arg_sep=ExtendedBool.INHERIT):
        '''Initializes an Option object with the specified parameters.

        Args:
            parent (CommandLine):
                The parent command line object, if any.

            option_strings (list of str):
                The list of option strings.

            metavar (str):
                The metavar for the option.

            help (str):
                The help message for the option.

            complete (tuple):
                The completion specification for the option.

            nosort (bool):
                Do not sort the completion suggestions.

            optional_arg (bool):
                Specifies if option's argument is optional.

            groups (list of str):
                Specify to which mutually exclusive groups this option
                belongs to.

            repeatable (ExtendedBool):
                Specifies if the option can be repeated.

            final (bool):
                If True, no more options are suggested after this one.

            hidden (bool):
                Specifies if this option is hidden.

            when (str):
                Specifies a condition for showing this option.

            capture (str):
                Specifies the variable name for capturing

            long_opt_arg_sep (str):
                Specifies which separators are used for delimiting a long
                option from its argument.

                If `space`, only the form `--option argument` is allowed.
                If `equals`, only the form `--option=argument` is allowed.
                If `both`, both forms are allowed.
                If `INHERIT`, the default setting from the config object is
                used.

        Returns:
            Option: The newly added Option object.
        '''

        validate_type(parent, (CommandLine, NoneType), 'parent')
        validate_type(option_strings, (list,), 'option_strings')
        validate_type(metavar, (str, NoneType), 'metavar')
        validate_type(help, (str, NoneType), 'help')
        validate_type(complete, (list, tuple, NoneType), 'complete')
        validate_type(groups, (list, NoneType), 'groups')

        if groups is not None:
            for index, group in enumerate(groups):
                if not isinstance(group, str):
                    raise CrazyTypeError(f'groups[{index}]', 'str', group)

        validate_type(optional_arg, (bool,), 'optional_arg')

        if not is_extended_bool(repeatable):
            raise CrazyTypeError('repeatable', 'ExtendedBool', repeatable)

        validate_type(final, (bool,), 'final')
        validate_type(hidden, (bool,), 'hidden')
        validate_type(nosort, (bool,), 'nosort')
        validate_type(when, (str, NoneType), 'when')
        validate_type(capture, (str, NoneType), 'capture')
        validate_type(long_opt_arg_sep, (str,), 'long_opt_arg_sep')

        if not option_strings:
            raise CrazyError('Empty option strings')

        for index, option_string in enumerate(option_strings):
            if not isinstance(option_string, str):
                raise CrazyTypeError(f'option_strings[{index}]', 'str', option_string)

            if not is_valid_option_string(option_string):
                raise CrazyError(f"Invalid option string: {option_string!r}")

        if capture is not None and not is_valid_variable_name(capture):
            raise CrazyError(f"Invalid variable name: {capture!r}")

        if metavar and not complete:
            raise CrazyError(f'Option {option_strings} has metavar set, but has no complete')

        if optional_arg and not complete:
            raise CrazyError(f'Option {option_strings} has optional_arg=True, but has no complete')

        if hidden is True and repeatable is True:
            raise CrazyError(f'Option {option_strings} has both hidden and repeatable set to True')

        if long_opt_arg_sep not in (ExtendedBool.INHERIT, 'space', 'equals', 'both'):
            raise CrazyError('Invalid value for `long_opt_arg_sep`')

        self.parent = parent
        self.option_strings = option_strings
        self.metavar = metavar
        self.help = help
        self.complete = complete
        self.nosort = nosort
        self.groups = groups
        self.optional_arg = optional_arg
        self.repeatable = repeatable
        self.final = final
        self.hidden = hidden
        self.when = when
        self.capture = capture
        self.long_opt_arg_sep = long_opt_arg_sep

    def get_option_strings(self):
        '''Returns the option strings associated with the Option object.

        Returns:
            list: A list of strings representing the option strings.
        '''

        return self.option_strings

    def get_option_strings_key(self, delimiter=' '):
        '''Returns a key with consisting of all option strings.'''

        return delimiter.join(sorted(self.option_strings))

    def get_short_option_strings(self):
        '''Returns the short option strings associated with the object.

        Returns:
            list: A list of short option strings ("-o").
        '''

        return list(filter(is_short_option_string, self.option_strings))

    def get_long_option_strings(self):
        '''Returns the long option strings associated with the object.

        Returns:
            list: A list of long option strings ("--option").
        '''

        return list(filter(is_long_option_string, self.option_strings))

    def get_old_option_strings(self):
        '''Returns the old-style option strings associated with the object.

        Returns:
            list: A list of old-style option strings ("-option").
        '''

        return list(filter(is_old_option_string, self.option_strings))

    def get_conflicting_options(self):
        '''Returns a list of conflicting options within the same mutually
        exclusive groups.

        Returns:
            list: A list of Option objects representing conflicting options.
        '''

        if not self.groups:
            return []
        r = []
        for group in self.groups:
            for option in self.parent.options:
                if option.groups is not None and group in option.groups:
                    if option not in r:
                        r.append(option)
        r.remove(self)
        return r

    def get_conflicting_option_strings(self):
        '''Returns a list of option strings conflicting with the current option
        within the same mutually exclusive groups.

        Returns:
            list: A list of option strings representing conflicting options.
        '''

        r = []
        for option in self.get_conflicting_options():
            r.extend(option.get_option_strings())
        return r

    def has_required_arg(self):
        '''Return True if the option takes a required argument.'''

        return self.complete is not None and self.optional_arg is False

    def has_optional_arg(self):
        '''Return True if the option takes an optional argument.'''

        return self.complete is not None and self.optional_arg is True

    def __eq__(self, other):
        return (
            isinstance(other, Option)                   and
            self.option_strings == other.option_strings and
            self.metavar        == other.metavar        and
            self.help           == other.help           and
            self.optional_arg   == other.optional_arg   and
            self.repeatable     == other.repeatable     and
            self.final          == other.final          and
            self.hidden         == other.hidden         and
            self.complete       == other.complete       and
            self.nosort         == other.nosort         and
            self.groups         == other.groups         and
            self.when           == other.when           and
            self.capture        == other.capture
        )

    def __repr__(self):
        return '{option_strings: %r, metavar: %r, help: %r}' % (
            self.option_strings, self.metavar, self.help)


class SubCommandsOption(Positional):
    '''Class holding subcommands of a commandline.'''

    def __init__(self, parent):
        self.subcommands = []

        super().__init__(
            parent,
            parent.get_highest_positional_num() + 1,
            metavar='command')

    def add_commandline_object(self, commandline):
        '''Add an existing commandline object as a subcommand.'''

        if self.get_subcommand_by_name(commandline.prog):
            raise CrazyError(f'Duplicate subcommand. {commandline.prog} already defined.')

        commandline.parent = self.parent
        self.subcommands.append(commandline)

    def add_commandline(self, name, help=''):
        '''Create a new commandline object and add it as a subcommand.'''

        if self.get_subcommand_by_name(name):
            raise CrazyError(f'Duplicate subcommand. {name} already defined.')

        commandline = CommandLine(name, help=help, parent=self.parent)
        self.subcommands.append(commandline)
        return commandline

    def get_choices(self, with_aliases=True):
        '''Return a mapping of command and its description.'''

        r = OrderedDict()
        for subcommand in self.subcommands:
            r[subcommand.prog] = subcommand.help
            if with_aliases:
                for alias in subcommand.aliases:
                    r[alias] = subcommand.help
        return r

    def get_subcommand_by_name(self, name):
        '''Return a subcommand by name or None if not found.'''

        for subcommand in self.subcommands:
            if subcommand.prog == name:
                return subcommand
        return None

    def __eq__(self, other):
        return (
            isinstance(other, SubCommandsOption)  and
            self.subcommands == other.subcommands and
            self.help        == other.help        and
            self.metavar     == other.metavar     and
            self.complete    == other.complete
        )

    def __repr__(self):
        return '{help: %r, subcommands %r}' % (self.help, self.subcommands)


class MutuallyExclusiveGroup:
    '''Helper class for adding mutually exclusive options.'''

    def __init__(self, parent, group):
        assert isinstance(parent, CommandLine)
        assert isinstance(group, str)

        self.parent = parent
        self.group = group

    def add(self, option_strings, **parameters):
        '''Creates and adds a new option.

        For a list of valid parameters, see `class Option`.
        '''

        if 'groups' in parameters:
            raise CrazyError('Paramter `groups` not allowed')

        if 'group' in parameters:
            raise CrazyError('Paramter `group` not allowed')

        return self.parent.add_option(
            option_strings,
            groups=[self.group],
            **parameters)

    def add_option(self, option):
        '''Adds an option object.'''

        option.parent = self.parent
        option.groups = [self.group]



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2025-2026 Benjamin Abendroth <braph93@gmx.de>

'''Module for positionals completion in Bash.'''

from . import algo
from . import bash_when
from .str_utils import indent


def _make_block(code):
    if not code:
        return '{\n  return 0;\n}'

    return '{\n%s\n  return 0;\n}' % indent(code, 2)


def _is_not_command_arg(positional):
    return (
        not positional.complete or
        positional.complete[0] != 'command_arg'
    )


def _generate_subcommand_positional(generator):
    cmds = generator.subcommands.get_choices().keys()
    complete = generator.completer.choices(generator.ctxt, [], cmds).get_code()
    return '(( ${#POSITIONALS[@]} == %d )) && %s' % (
        generator.subcommands.get_positional_num(),
        _make_block(complete))


def _get_positional_condition(positional):
    op = '=='
    if positional.repeatable:
        op = '>='
    return '${#POSITIONALS[@]} %s %s' % (op, positional.get_positional_num())


def _generate_positionals_with_when(generator):
    code = []

    positionals = filter(_is_not_command_arg, generator.positionals)

    for positional in positionals:
        condition = _get_positional_condition(positional)
        r = '(( %s )) && ' % condition
        if positional.when:
            r += '%s && ' % bash_when.generate_when_conditions(
                generator.commandline,
                generator.variable_manager,
                positional.when)
        r += '%s' % _make_block(generator.complete_option(positional))

        code.append(r)

    return '\n\n'.join(code)


def _generate_positionals_without_when(generator):
    code = []

    positionals = filter(_is_not_command_arg, generator.positionals)

    grouped_by_complete = algo.group_by(
        positionals, generator.complete_option)

    for complete, positionals in grouped_by_complete.items():
        conditions = [_get_positional_condition(p) for p in positionals]
        r = '(( %s )) && ' % ' || '.join(conditions)
        r += '%s' % _make_block(complete)
        code.append(r)

    return '\n\n'.join(code)


def generate(generator):
    '''Generate code for completing positionals.'''

    code = []

    has_when = any(p.when for p in generator.positionals)

    if has_when:
        code += [_generate_positionals_with_when(generator)]
    else:
        code += [_generate_positionals_without_when(generator)]

    if generator.subcommands:
        code += [_generate_subcommand_positional(generator)]

    return '\n\n'.join(c for c in code if c)
